"""
Keyed collection proxies.
"""

from collections.abc import Iterable, Iterator
from typing import Any, override

from betty.collection.keyed import KeyedCollection, MutableKeyedCollection


class _KeyedCollectionProxy[
    KeyT,
    ResolvableKeyT,
    ValueT,
    KeyedCollectionT: KeyedCollection,
](KeyedCollection[KeyT, ResolvableKeyT, ValueT]):
    def __init__(self, upstream: KeyedCollectionT, /):
        self._upstream = upstream

    @override
    def __len__(self) -> int:
        return len(self._upstream)

    @override
    def __iter__(self) -> Iterator[ValueT]:
        return iter(self._upstream)

    @override
    def __contains__(self, key: Any) -> bool:
        return key in self._upstream

    @override
    def __getitem__(self, key: KeyT | ResolvableKeyT) -> ValueT:
        return self._upstream[key]

    @override
    def keys(self) -> Iterable[KeyT]:
        return self._upstream.keys()


class KeyedCollectionProxy[KeyT, ResolvableKeyT, ValueT](
    _KeyedCollectionProxy[
        KeyT, ResolvableKeyT, ValueT, KeyedCollection[KeyT, ResolvableKeyT, ValueT]
    ]
):
    """
    A keyed collection proxy.
    """


class MutableKeyedCollectionProxy[KeyT, ResolvableKeyT, ValueT, ResolvableValueT](
    _KeyedCollectionProxy[
        KeyT,
        ResolvableKeyT,
        ValueT,
        MutableKeyedCollection[KeyT, ResolvableKeyT, ValueT, ResolvableValueT],
    ],
    MutableKeyedCollection[KeyT, ResolvableKeyT, ValueT, ResolvableValueT],
):
    """
    A mutable keyed collection proxy.
    """

    @override
    def remove(self, *keys: KeyT | ResolvableKeyT) -> None:
        self._upstream.remove(*keys)

    @override
    def __delitem__(self, key: KeyT | ResolvableKeyT) -> None:
        del self._upstream[key]

    @override
    def clear(self) -> None:
        self._upstream.clear()

    @override
    def add(self, *values: ValueT | ResolvableValueT) -> None:
        self._upstream.add(*values)



"""
Jobs.
"""

from __future__ import annotations

from asyncio import gather
from collections import defaultdict
from typing import TYPE_CHECKING, final, override

from aiohttp import ClientError, ClientSession
from lxml.html import HtmlElement, document_fromstring

from betty.job import Job
from betty.locale.localizable.static import StaticTranslations
from betty.media_type import InvalidMediaType, MediaType

if TYPE_CHECKING:
    from collections.abc import Iterable

    from babel import Locale

    from betty.ancestry.link import Link
    from betty.job.scheduler import Scheduler
    from betty.locale.localizable import StaticTranslationsMapping
    from betty.locale.localize import Localizer


@final
class PopulateLink(Job):
    """
    Populate a link with information from its URL.
    """

    def __init__(
        self, link: Link, *, http_client: ClientSession, localizers: Iterable[Localizer]
    ):
        super().__init__(self.id_for(link), priority=True)
        self._link = link
        self._http_client = http_client
        self._localizers = localizers

    @classmethod
    def id_for(cls, link: Link) -> str:
        """
        Get the job ID.
        """
        return f"populate-link:{link.id}"

    @override
    async def do(self, scheduler: Scheduler, /) -> None:
        if self._link.has_label and self._link.description:
            return

        urls = StaticTranslations.resolve(self._link.url, self._localizers)
        urls_to_locales = defaultdict(set)
        for locale, url in urls.translations.items():
            urls_to_locales[url].add(locale)
        labels: StaticTranslationsMapping = {}
        descriptions: StaticTranslationsMapping = {}
        await gather(
            *(
                self._populate_link_from_url(
                    url,
                    [localizer.locale for localizer in self._localizers],
                    labels,
                    descriptions,
                )
                for url in urls_to_locales
            )
        )
        if not self._link.has_label and labels:
            self._link.label = StaticTranslations(labels)
        if not self._link.description and descriptions:
            self._link.description = StaticTranslations(descriptions)

    async def _populate_link_from_url(
        self,
        url: str,
        locales: Iterable[Locale],
        labels: StaticTranslationsMapping,
        descriptions: StaticTranslationsMapping,
    ) -> None:
        try:
            response = await self._http_client.get(url)
        except ClientError:
            return
        try:
            content_type = MediaType(response.headers["Content-Type"])
        except InvalidMediaType:
            return

        if (content_type.type, content_type.subtype, content_type.suffix) not in (
            ("text", "html", None),
            ("application", "xhtml", "+xml"),
        ):
            return

        document = document_fromstring(await response.text())
        if not self._link.has_label:
            title = self._extract_html_title(document)
            if title is not None:
                for locale in locales:
                    labels[locale] = title
        if not self._link.description:
            description = self._extract_html_meta_description(document)
            if description is not None:
                for locale in locales:
                    descriptions[locale] = description

    def _extract_html_title(self, document: HtmlElement) -> str | None:
        head = document.find("head")
        if head is None:
            return None
        title = head.find("title")
        if title is None:
            return None
        return title.text

    def _extract_html_meta_description(self, document: HtmlElement) -> str | None:
        head = document.find("head")
        if head is None:
            return None
        metas = head.findall("meta")
        for attr_name, attr_value in (
            ("name", "description"),
            ("property", "og:description"),
        ):
            for meta in metas:
                if meta.get(attr_name, None) == attr_value:
                    return meta.get("content", None)
        return None



"""Privatize people and associated data by determining if they are likely to be alive."""

from __future__ import annotations

from typing import TYPE_CHECKING, Self, final, override

from betty.extension import Extension, ExtensionDefinition
from betty.extension.deriver import Deriver
from betty.extension.deriver.jobs import DeriveAncestry
from betty.extension.privatizer.jobs import PrivatizeAncestry
from betty.locale.localizable.gettext import _
from betty.project import Project
from betty.project.load import PostLoader
from betty.service.factory import Manufacturable
from betty.service.requirement.project import require_project

if TYPE_CHECKING:
    from betty.job.scheduler import Scheduler


@final
@ExtensionDefinition(
    "privatizer",
    label="Privatizer",
    description=_(
        "Determine if people can be proven to have died. If not, mark them and their associated entities private."
    ),
    comes_after={Deriver},
)
class Privatizer(PostLoader, Manufacturable, Extension[Project]):
    """
    .. plugin:: extension:privatizer.

    Entities in Betty have privacy. This is a *ternary* property, with the following possible values:

    public (``entity.privacy = betty.privacy.Privacy.PUBLIC``)
        The entity will be included when publishing your ancestry. The privacy **should not** be changed.
    private (``entity.privacy = betty.privacy.Privacy.PRIVATE``)
        The entity will not be included when publishing your ancestry. The privacy **should not** be changed.
    undetermined (``entity.privacy = betty.privacy.Privacy.UNDETERMINED``)
        The entity is public, but its privacy **may** be determined or changed at will.

    The following entities are processed by the Privatizer. They are marked *private* except if any of the following
    conditions are met:

    People
      People are considered dead past the *lifetime threshold*, which
      :py:const:`defaults to 123 years <betty.project.data.DEFAULT_LIFETIME_THRESHOLD>`, but can be changed in your
      project's :py:class:`configuration <betty.project.data.ProjectConfiguration>`.

      * The person has an end-of-life event, such as a death, final disposition, or will.
      * Any event that was at least the *lifetime threshold* ago.
      * For every person *n* generation(s) before this person, if that person has an end-of-life event at least *n* *
        *lifetime threshold* ago.
      * For every person *n* generation(s) before this person, if that person has any event that was at least (*n* + 1) *
        *lifetime threshold* ago.
      * For every descendant if that person has any event that was at least *lifetime threshold* ago.

      If the Privatizer determines a person private, it will also privatize any events, citations, and files associated
      with that person.

    File
      Any citations associated with private files will be privatized.

    Event
      Any citations and files associated with private events will be privatized.

    Citation
      The source and any files associated with private citations will be privatized.

    Source
      Any files associated with private sources will be privatized.

    """

    def __init__(self, *, project: Project):
        super().__init__()
        self._project = project

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        return cls(project=project)

    @override
    async def post_load(self, scheduler: Scheduler) -> None:
        await scheduler.add(
            PrivatizeAncestry(
                dependencies={DeriveAncestry.id_for()}
                if Deriver.plugin().id in await self._project.extensions
                else set(),
                project=self._project,
            )
        )



"""
Locale API errors.
"""

from __future__ import annotations

from typing import final

from babel import Locale
from babel.localedata import locale_identifiers

from betty.exception import HumanFacingException
from betty.locale import to_language_tag
from betty.locale.localizable.gettext import _
from betty.locale.localizable.markup import Paragraph, do_you_mean


class LocaleError(HumanFacingException, Exception):
    """
    A locale API error.
    """


@final
class InvalidLocale(LocaleError):
    """
    Raised when a value is not a valid locale.
    """

    def __init__(self, invalid_locale: str, /) -> None:
        super().__init__(
            _('"{invalid_locale}" is not a valid IETF BCP 47 language tag.').format(
                invalid_locale=invalid_locale
            )
        )


@final
class UnknownLocale(LocaleError):
    """
    Raised when a locale is not known by the system.
    """

    def __init__(self, locale: str, /) -> None:
        locale_chars = {char for char in locale[: locale.find("-")] if char.isalpha()}
        available_locales = sorted(
            to_language_tag(Locale.parse(identifier))
            for identifier in locale_identifiers()
            if set(identifier[: identifier.find("_")]) & locale_chars
        )
        super().__init__(
            Paragraph(
                _("Locale {locale} is not known by your system.").format(locale=locale),
                do_you_mean(*available_locales),
            )
        )



"""
Generic plugin API errors.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from betty.exception import HumanFacingException
from betty.locale.localizable.gettext import _
from betty.locale.localizable.markup import Paragraph, do_you_mean
from betty.plugin import (
    PluginDefinition,
    ResolvableId,
    resolve_id,
)

if TYPE_CHECKING:
    from collections.abc import Iterable

    from betty.machine_name import ResolvableMachineName


class PluginError(Exception):
    """
    Any error originating from the Plugin API.
    """


class PluginTypeNotFound(PluginError, HumanFacingException):
    """
    Raised when a plugin type cannot be found.
    """

    def __init__(
        self,
        plugin_type_not_found: ResolvableMachineName,
        available_plugin_types: Iterable[ResolvableMachineName],
        /,
    ):
        super().__init__(
            Paragraph(
                _('Cannot find the "{plugin_type}" plugin type.').format(
                    plugin_type=plugin_type_not_found
                ),
                do_you_mean(
                    *[
                        f'"{available_plugin_type}"'
                        for available_plugin_type in available_plugin_types
                    ]
                ),
            )
        )


class PluginNotFound(PluginError, HumanFacingException):
    """
    Raised when a plugin cannot be found.
    """

    def __init__[PluginDefinitionT: PluginDefinition](
        self,
        plugin_type: type[PluginDefinitionT],
        plugin_not_found: ResolvableMachineName,
        available_plugins: Iterable[ResolvableId[PluginDefinitionT]],
        /,
    ):
        super().__init__(
            Paragraph(
                _('Cannot find the "{plugin_id}" {plugin_type} plugin.').format(
                    plugin_type=plugin_type.type().label, plugin_id=plugin_not_found
                ),
                do_you_mean(
                    *[
                        f'"{resolve_id(available_plugin)}"'
                        for available_plugin in available_plugins
                    ]
                ),
            )
        )



"""
Data types to describe information sources.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final, override

from betty.ancestry.date import HasDate
from betty.ancestry.has_file_references import HasFileReferences
from betty.ancestry.has_links import HasLinks
from betty.ancestry.has_notes import HasNotes
from betty.json.linked_data import JsonLdObject, dump_context
from betty.locale.localizable.gettext import _, ngettext
from betty.locale.localizable.linked_data import dump_linked_data
from betty.locale.localizable.property import LocalizableProperty
from betty.locale.localizable.static.schema import StaticTranslationsSchema
from betty.model import Entity, EntityDefinition
from betty.model.association import (
    BidirectionalToManySingleType,
    BidirectionalToZeroOrOne,
    ToManyAssociates,
    ToZeroOrOneAssociate,
)
from betty.privacy import HasPrivacy, Privacy, is_public, merge_privacies
from betty.property import Optional

if TYPE_CHECKING:
    from collections.abc import MutableSequence

    from betty.ancestry.citation import Citation  # noqa: F401
    from betty.ancestry.file_reference import FileReference
    from betty.ancestry.link import Link
    from betty.ancestry.note import Note
    from betty.date import ResolvableDate
    from betty.locale.localizable import Localizable, ResolvableLocalizable
    from betty.portable import PortableMapping
    from betty.project import Project


@final
@EntityDefinition(
    "source",
    label=_("Source"),
    label_plural=_("Sources"),
    label_countable=ngettext("{count} source", "{count} sources"),
)
class Source(HasDate, HasFileReferences, HasNotes, HasLinks, HasPrivacy, Entity):
    """
    .. plugin:: entity:source.
    """

    name = Optional(LocalizableProperty(label=_("Name")))
    """
    The source's name.
    """

    author = Optional(LocalizableProperty(label=_("Author")))
    """
    The source's author.
    """

    publisher = Optional(LocalizableProperty(label=_("Publisher")))
    """
    The source's publisher.
    """

    contained_by = BidirectionalToZeroOrOne["Source", "Source"](
        "betty.ancestry.source:Source",
        "contains",
        label=_("Contained by"),
        description=_("Another source this source may be contained by"),
    )
    """
    Another source this source may be contained by
    """

    contains = BidirectionalToManySingleType["Source", "Source"](
        "betty.ancestry.source:Source",
        "contained_by",
        label=_("Contains"),
        description=_("Other sources this source may contain"),
    )
    """
    Other sources this source may contain
    """

    citations = BidirectionalToManySingleType["Source", "Citation"](
        "betty.ancestry.citation:Citation",
        "source",
        label=_("Citations"),
        description=_("The citations referencing this source"),
    )
    """
    The citations referencing this source
    """

    def __init__(
        self,
        name: ResolvableLocalizable | None = None,
        *,
        id: str | None = None,  # noqa: A002
        author: ResolvableLocalizable | None = None,
        publisher: ResolvableLocalizable | None = None,
        contained_by: ToZeroOrOneAssociate[Source] = None,
        contains: ToManyAssociates[Source] | None = None,
        notes: ToManyAssociates[Note] | None = None,
        date: ResolvableDate | None = None,
        file_references: ToManyAssociates[FileReference] | None = None,
        links: MutableSequence[Link] | None = None,
        privacy: Privacy | None = None,
    ):
        super().__init__(
            id,
            notes=notes,
            date=date,
            file_references=file_references,
            links=links,
            privacy=privacy,
        )
        self.name = name
        self.author = author
        self.publisher = publisher
        if contained_by is not None:
            self.contained_by = contained_by
        if contains is not None:
            self.contains = contains

    @override
    def _get_effective_privacy(self) -> Privacy:
        privacy = super()._get_effective_privacy()
        if self.contained_by:
            return merge_privacies(privacy, self.contained_by)
        return privacy

    @override
    @property
    def label(self) -> Localizable:
        return self.name if self.name else super().label

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project, /) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        static_translations_schema = StaticTranslationsSchema()
        schema.add_property("author", static_translations_schema, False)
        schema.add_property("name", static_translations_schema, False)
        schema.add_property("publisher", static_translations_schema, False)
        return schema

    @override
    async def dump_linked_data(self, project: Project, /) -> PortableMapping:
        portable = await super().dump_linked_data(project)
        portable["@type"] = "https://schema.org/Thing"
        dump_context(portable, name="https://schema.org/name")
        if is_public(self):
            public_localizers = await project.public_localizers
            if self.author is not None:
                portable["author"] = dump_linked_data(
                    self.author, localizers=public_localizers
                )
            if self.name is not None:
                portable["name"] = dump_linked_data(
                    self.name, localizers=public_localizers
                )
            if self.publisher is not None:
                portable["publisher"] = dump_linked_data(
                    self.publisher, localizers=public_localizers
                )
        return portable



"""
Data types for citations.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final, override

from betty.ancestry.date import HasDate
from betty.ancestry.has_file_references import HasFileReferences
from betty.ancestry.has_links import HasLinks
from betty.ancestry.source import Source
from betty.locale.localizable.gettext import _, ngettext
from betty.locale.localizable.linked_data import dump_linked_data
from betty.locale.localizable.property import LocalizableProperty
from betty.locale.localizable.static.schema import StaticTranslationsSchema
from betty.model import EntityDefinition
from betty.model.association import (
    BidirectionalToManyMultipleTypes,
    BidirectionalToOne,
    ToManyAssociates,
    ToOneAssociate,
)
from betty.privacy import HasPrivacy, Privacy, is_public, merge_secondary_privacies
from betty.property import Optional

if TYPE_CHECKING:
    from betty.ancestry.file_reference import FileReference
    from betty.ancestry.has_citations import HasCitations
    from betty.date import ResolvableDate
    from betty.json.linked_data import JsonLdObject
    from betty.locale.localizable import Localizable, ResolvableLocalizable
    from betty.portable import PortableMapping
    from betty.project import Project


@final
@EntityDefinition(
    "citation",
    label=_("Citation"),
    label_plural=_("Citations"),
    label_countable=ngettext("{count} citation", "{count} citations"),
)
class Citation(HasDate, HasFileReferences, HasPrivacy, HasLinks):
    """
    .. plugin:: entity:citation.
    """

    location = Optional(LocalizableProperty(label=_("Location")))
    """
    The location within the source this citation references.
    """

    facts = BidirectionalToManyMultipleTypes["Citation", "HasCitations"](
        "betty.ancestry.has_citations:HasCitations",
        "citations",
        label=_("Facts"),
        description=_(
            "The other entities that reference these citations to back up their claims."
        ),
    )
    """
    The other entities that reference these citations to back up their claims.
    """

    source = BidirectionalToOne["Citation", Source](
        "betty.ancestry.source:Source",
        "citations",
        label=_("Source"),
        description=_("The source this citation references."),
    )
    """
    The source this citation references.
    """

    def __init__(
        self,
        *,
        source: ToOneAssociate[Source],
        id: str | None = None,  # noqa: A002
        facts: ToManyAssociates[HasCitations] | None = None,
        location: ResolvableLocalizable | None = None,
        date: ResolvableDate | None = None,
        file_references: ToManyAssociates[FileReference] | None = None,
        privacy: Privacy | None = None,
    ):
        super().__init__(
            id,
            date=date,
            file_references=file_references,
            privacy=privacy,
        )
        if facts is not None:
            self.facts = facts
        self.location = location
        self.source = source

    @override
    def _get_effective_privacy(self) -> Privacy:
        return merge_secondary_privacies(super()._get_effective_privacy(), self.source)

    @override
    @property
    def label(self) -> Localizable:
        return self.location or super().label

    @override
    async def dump_linked_data(self, project: Project, /) -> PortableMapping:
        portable = await super().dump_linked_data(project)
        portable["@type"] = "https://schema.org/Thing"
        if is_public(self) and self.location is not None:
            portable["location"] = dump_linked_data(
                self.location, localizers=await project.public_localizers
            )
        return portable

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project, /) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        schema.add_property(
            "location",
            StaticTranslationsSchema(),
            False,
        )
        return schema



"""
Test utilities for :py:mod:`betty.job.executor`.
"""

from __future__ import annotations

from asyncio import sleep
from typing import TYPE_CHECKING

import pytest

from betty.job import Context
from betty.job.scheduler import Cancelled, ScheduledJobBatch, Scheduler
from betty.test_utils.job.scheduler import StaticScheduler

if TYPE_CHECKING:
    from collections.abc import Callable

    from betty.job.executor import Executor


async def _sleep() -> None:
    await sleep(999)


class ExecutorTestBase:
    """
    A base class for testing :py:class:`betty.job.executor.Executor` implementations.
    """

    @pytest.fixture
    async def new_sut(self) -> Callable[[Scheduler], Executor]:
        """
        Provide the systems under test.
        """
        raise NotImplementedError

    async def test___aexit___with_scheduler_cancelled(
        self, new_sut: Callable[[Scheduler], Executor]
    ) -> None:
        """
        Tests :py:meth:`betty.job.executor.Executor.__aexit__` implementations.
        """
        scheduler = StaticScheduler(Context(), [Cancelled()])
        async with new_sut(scheduler):
            pass

    async def test___aexit___with_scheduler_completed(
        self, new_sut: Callable[[Scheduler], Executor]
    ) -> None:
        """
        Tests :py:meth:`betty.job.executor.Executor.__aexit__` implementations.
        """
        scheduler = StaticScheduler(Context(), [])
        async with new_sut(scheduler):
            pass

    async def test___aexit___with_arbitrary_exception(
        self, new_sut: Callable[[Scheduler], Executor]
    ) -> None:
        """
        Tests :py:meth:`betty.job.executor.Executor.__aexit__` implementations.
        """
        exception = RuntimeError()
        scheduler = StaticScheduler(Context(), [_sleep for _ in range(999)])
        with pytest.raises(type(exception)) as exc_info:
            async with new_sut(scheduler):
                raise exception
        assert exc_info.value is exception

    async def test_start(self, new_sut: Callable[[Scheduler], Executor]) -> None:
        """
        Tests :py:meth:`betty.job.executor.Executor.start` implementations.
        """
        scheduler = StaticScheduler(Context(), [])
        sut = new_sut(scheduler)
        await sut.start()
        await sut.complete()

    async def test_start__when_started(
        self, new_sut: Callable[[Scheduler], Executor]
    ) -> None:
        """
        Tests :py:meth:`betty.job.executor.Executor.start` implementations.
        """
        scheduler = StaticScheduler(Context(), [])
        sut = new_sut(scheduler)
        await sut.start()
        await sut.start()
        await sut.complete()

    @pytest.mark.parametrize(
        "batch_count",
        [
            1,
            999,
        ],
    )
    async def test_complete(
        self, batch_count: int, new_sut: Callable[[Scheduler], Executor]
    ) -> None:
        """
        Tests :py:meth:`betty.job.executor.Executor.complete` implementations.
        """
        actual = []

        def _build_batch(batch_index: int) -> ScheduledJobBatch:
            async def _batch() -> None:
                actual.append(f"batch:{batch_index}")

            return _batch

        expected = [f"batch:{batch_index}" for batch_index in range(batch_count)]

        scheduler = StaticScheduler(
            Context(), [_build_batch(batch_index) for batch_index in range(batch_count)]
        )
        sut = new_sut(scheduler)
        await sut.start()
        await sut.complete()
        assert actual == expected

    async def test_complete__when_not_started(
        self, new_sut: Callable[[Scheduler], Executor]
    ) -> None:
        """
        Tests :py:meth:`betty.job.executor.Executor.complete` implementations.
        """
        scheduler = StaticScheduler(Context(), [])
        sut = new_sut(scheduler)
        await sut.complete()

    async def test_complete__when_completed(
        self, new_sut: Callable[[Scheduler], Executor]
    ) -> None:
        """
        Tests :py:meth:`betty.job.executor.Executor.complete` implementations.
        """
        scheduler = StaticScheduler(Context(), [])
        sut = new_sut(scheduler)
        await sut.start()
        await sut.complete()
        await sut.complete()

    async def test_cancel(self, new_sut: Callable[[Scheduler], Executor]) -> None:
        """
        Tests :py:meth:`betty.job.executor.Executor.cancel` implementations.
        """
        scheduler = StaticScheduler(Context(), [_sleep for _ in range(999)])
        sut = new_sut(scheduler)
        await sut.start()
        await sut.cancel()

    async def test_cancel__when_not_started(
        self, new_sut: Callable[[Scheduler], Executor]
    ) -> None:
        """
        Tests :py:meth:`betty.job.executor.Executor.cancel` implementations.
        """
        scheduler = StaticScheduler(Context(), [])
        sut = new_sut(scheduler)
        await sut.cancel()

    async def test_cancel__when_cancelled(
        self, new_sut: Callable[[Scheduler], Executor]
    ) -> None:
        """
        Tests :py:meth:`betty.job.executor.Executor.cancel` implementations.
        """
        scheduler = StaticScheduler(Context(), [])
        sut = new_sut(scheduler)
        await sut.start()
        await sut.cancel()
        await sut.cancel()



"""
Manage documents.

A document is a singular file, such as an HTML page or a JSON file.
"""

from __future__ import annotations

from collections import defaultdict
from collections.abc import (
    Iterable,
    Iterator,
    Mapping,
    MutableMapping,
    MutableSequence,
    Sized,
)
from threading import Lock
from typing import (
    TYPE_CHECKING,
    Any,
    Self,
    final,
    override,
)

from betty.json.linked_data import LinkedDataDumpable
from betty.locale.localize import DEFAULT_LOCALIZER, Localizer
from betty.media_type.media_types import HTML
from betty.plugin import ResolvableId, resolve_id
from betty.portable import PortableMapping

if TYPE_CHECKING:
    from betty.ancestry.citation import Citation
    from betty.job import Context
    from betty.locale.localizable import Localizable
    from betty.machine_name import MachineName
    from betty.model import Entity
    from betty.project import Project

type DocumentVars = Mapping[str, Any]


@final
class Document:
    """
    A document.
    """

    def __init__(
        self,
        resource: object = None,
        resource_url: object = None,
        *,
        breadcrumbs: Breadcrumbs | None = None,
        citer: Citer | None = None,
        entity_contexts: EntityContexts | None = None,
        context: Context | None = None,
        localizer: Localizer | None = None,
        title: Localizable | None = None,
        **vars: Any,  # noqa: A002
    ):
        self._resource = resource
        self._resource_url = resource_url
        self._entity_contexts = entity_contexts if entity_contexts else EntityContexts()
        self._context = context
        self._localizer = localizer if localizer else DEFAULT_LOCALIZER
        self._title = title
        self._vars = vars
        self._breadcrumbs = Breadcrumbs() if breadcrumbs is None else breadcrumbs
        self._citer = Citer() if citer is None else citer

    @property
    def breadcrumbs(self) -> Breadcrumbs:
        """
        The breadcrumbs.
        """
        return self._breadcrumbs

    @property
    def citer(self) -> Citer:
        """
        The citer.
        """
        return self._citer

    @property
    def entity_contexts(self) -> EntityContexts:
        """
        The entity contexts.
        """
        return self._entity_contexts

    @property
    def context(self) -> Context | None:
        """
        The job context.
        """
        return self._context

    @property
    def localizer(self) -> Localizer:
        """
        The localizer.
        """
        return self._localizer

    @property
    def resource(self) -> object:
        """
        The resource itself.
        """
        return self._resource

    @property
    def resource_url(self) -> object:
        """
        The URL-generatable version of the resource itself.

        This may be the resource itself or a completely different type of value.
        """
        return self._resource_url

    @property
    def title(self) -> Localizable | None:
        """
        The human-readable title.
        """
        return self._title

    def copy(
        self,
        **vars: object,  # noqa: A002
    ) -> Self:
        """
        Create a copy of this document, with the given fields added.
        """
        return type(self)(
            **{
                **self._vars,
                "resource": self._resource,
                "resource_url": self._resource_url,
                "breadcrumbs": self._breadcrumbs,
                "citer": self._citer,
                "entity_contexts": self._entity_contexts,
                "context": self._context,
                "localizer": self._localizer,
                "title": self._title,
                **vars,
            },  # ty:ignore[invalid-argument-type]
        )

    def __getitem__(self, var: str) -> object:
        return self._vars[var]

    def __setitem__(self, var: str, value: object) -> None:
        self._vars[var] = value

    def __contains__(self, var: str) -> bool:
        return var in self._vars

    @override
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Document):
            return NotImplemented
        return (
            self._resource == other._resource
            and self._resource_url == other._resource_url
            and self._entity_contexts == other._entity_contexts
            and self._context == other._context
            and self._localizer == other._localizer
            and self._title == other._title
            and self._breadcrumbs == other._breadcrumbs
            and self._citer == other._citer
            and self._vars == other._vars
        )


class DocumentProvider:
    """
    Provide new documents.
    """

    def new_document_vars(self) -> DocumentVars:
        """
        Create new variables for a new :py:class:`betty.document.Document`.

        Keys are the variable names, and values are variable values.
        """
        return {}


@final
class Breadcrumb(LinkedDataDumpable[PortableMapping]):
    """
    A breadcrumb.
    """

    def __init__(self, label: str, resource: object | None, /):
        self._label = label
        self._resource_url = resource

    @property
    def label(self) -> str:
        """
        The localized, human-readable label.
        """
        return self._label

    @property
    def resource_url(self) -> object | None:
        """
        The resource URL.
        """
        return self._resource_url

    @override
    async def dump_linked_data(self, project: Project, /) -> PortableMapping:
        portable: PortableMapping = {
            "@type": "ListItem",
            "name": self._label,
        }
        if self._resource_url is not None:
            url_generator = await project.url_generator
            portable["item"] = url_generator.generate(
                self._resource_url, absolute=True, media_type=HTML
            )
        return portable


@final
class Breadcrumbs(LinkedDataDumpable[PortableMapping], Iterable[Breadcrumb], Sized):
    """
    A trail of navigational breadcrumbs.
    """

    def __init__(self):
        self._breadcrumbs: MutableSequence[Breadcrumb] = []

    @override
    def __iter__(self) -> Iterator[Breadcrumb]:
        return iter(self._breadcrumbs)

    @override
    def __len__(self) -> int:
        return len(self._breadcrumbs)

    def append(self, label: str, resource_url: object | None = None, /) -> None:
        """
        Append a breadcrumb to the trail.
        """
        self._breadcrumbs.append(Breadcrumb(label, resource_url))

    @override
    async def dump_linked_data(self, project: Project, /) -> PortableMapping:
        if not self._breadcrumbs:
            return {}
        return {
            "@context": "https://schema.org",
            "@type": "BreadcrumbList",
            "itemListElement": [
                {
                    "position": position,
                    **await breadcrumb.dump_linked_data(project),
                }
                for position, breadcrumb in enumerate(self._breadcrumbs, 1)
            ],
        }


class Citer:
    """
    Track citations when they are first used.
    """

    __slots__ = "_lock", "_cited"

    def __init__(self):
        self._lock = Lock()
        self._cited: MutableSequence[Citation] = []

    def __iter__(self) -> enumerate[Citation]:
        return enumerate(self._cited, 1)

    def __len__(self) -> int:
        return len(self._cited)

    def cite(self, citation: Citation, /) -> int:
        """
        Reference a citation.

        :returns: The citation's sequential reference number.
        """
        with self._lock:
            if citation not in self._cited:
                self._cited.append(citation)
            return self._cited.index(citation) + 1


class EntityContexts:
    """
    Track the current entity contexts.

    To allow templates to respond to their environment, this class allows
    our templates to set and get one entity per entity type for the current context.

    Use cases include rendering an entity label as plain text if the template is in
    that entity's context, but as a hyperlink if the template is not in the entity's
    context.
    """

    def __init__(self, *entities: Entity) -> None:
        self._contexts: MutableMapping[MachineName, Entity | None] = defaultdict(
            lambda: None
        )
        for entity in entities:
            self._contexts[entity.plugin().id] = entity

    def __getitem__(self, entity_type: ResolvableId) -> Entity | None:
        return self._contexts[resolve_id(entity_type)]

    def __call__(self, *entities: Entity) -> EntityContexts:
        """
        Create a new context with the given entities.
        """
        updated_contexts = EntityContexts(
            *(entity for entity in self._contexts.values() if entity is not None)
        )
        for entity in entities:
            updated_contexts._contexts[entity.plugin().id] = entity
        return updated_contexts



"""
Data indicators.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, final, override

if TYPE_CHECKING:
    import pathlib


class Indicator(ABC):
    """
    Describe a location of a piece of data.
    """

    @abstractmethod
    def format(self) -> str:
        """
        Format the indicator to a string.
        """


@final
class AnyIndex(Indicator):
    """
    A sequence item indicator.
    """

    @override
    def format(self) -> str:
        return "[]"


@final
class AnyKey(Indicator):
    """
    A mapping item indicator.
    """

    @override
    def format(self) -> str:
        return "{}"


@final
class Path(Indicator):
    """
    A file on disk.
    """

    def __init__(self, path: pathlib.Path, /):
        self._path = path.resolve().absolute()

    @override
    def format(self) -> str:
        return str(self._path)


@final
class Url(Indicator):
    """
    A URL.
    """

    def __init__(self, url: str, /):
        self._url = url

    @override
    def format(self) -> str:
        return self._url



"""
Entity collections.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections import defaultdict
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, cast, override

from betty.functools import unique
from betty.model import Entity, EntityDefinition

if TYPE_CHECKING:
    from collections.abc import (
        Iterable,
        Iterator,
        MutableMapping,
        MutableSequence,
        Sequence,
    )

    from ty_extensions import Intersection

    from betty.machine_name import ResolvableMachineName


class EntityCollection[TargetT = Entity](ABC):
    """
    Provide a collection of entities.

    To test your own subclasses, use :py:class:`betty.test_utils.model.collections.EntityCollectionTestBase`.
    """

    def _on_add(  # noqa: B027
        self,
        *entities: Intersection[TargetT, Entity],
    ) -> None:
        pass

    def _on_remove(  # noqa: B027
        self,
        *entities: Intersection[TargetT, Entity],
    ) -> None:
        pass

    @property
    def view(self) -> Sequence[Intersection[TargetT, Entity]]:
        """
        A view of the entities at the time of calling.
        """
        return [*self]

    @abstractmethod
    def add(self, *entities: Intersection[TargetT, Entity]) -> None:
        """
        Add the given entities.
        """

    @abstractmethod
    def remove(self, *entities: Intersection[TargetT, Entity]) -> None:
        """
        Remove the given entities.
        """

    def replace(self, *entities: Intersection[TargetT, Entity]) -> None:
        """
        Replace all entities with the given ones.
        """
        self.remove(*(entity for entity in self if entity not in entities))
        self.add(*entities)

    @abstractmethod
    def clear(self) -> None:
        """
        Clear all entities from the collection.
        """

    @abstractmethod
    def __iter__(self) -> Iterator[Intersection[TargetT, Entity]]:
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def __delitem__(self, key: Intersection[TargetT, Entity]) -> None:
        pass

    @abstractmethod
    def __contains__(self, value: Any) -> bool:
        pass

    def _known(
        self, *entities: Intersection[TargetT, Entity]
    ) -> Iterable[Intersection[TargetT, Entity]]:
        for entity in unique(entities):
            if entity in self:
                yield entity

    def _unknown(
        self, *entities: Intersection[TargetT, Entity]
    ) -> Iterable[Intersection[TargetT, Entity]]:
        for entity in unique(entities):
            if entity not in self:
                yield entity


class SingleTypeEntityCollection[TargetT = Entity](EntityCollection[TargetT]):
    """
    Collect entities of a single type.
    """

    def __init__(self, *entities: Intersection[TargetT, Entity]):
        super().__init__()
        self._entities: MutableSequence[Intersection[TargetT, Entity]] = [*entities]

    @override
    def add(self, *entities: Intersection[TargetT, Entity]) -> None:
        added_entities = [*self._unknown(*entities)]
        for entity in added_entities:
            self._entities.append(entity)
        if added_entities:
            self._on_add(*added_entities)

    @override
    def remove(self, *entities: Intersection[TargetT, Entity]) -> None:
        removed_entities = [*self._known(*entities)]
        for entity in removed_entities:
            self._entities.remove(entity)
        if removed_entities:
            self._on_remove(*removed_entities)

    @override
    def clear(self) -> None:
        self.remove(*self)

    @override
    def __iter__(self) -> Iterator[Intersection[TargetT, Entity]]:
        return self._entities.__iter__()

    @override
    def __len__(self) -> int:
        return len(self._entities)

    def __getitem__(self, entity_id: str) -> Intersection[TargetT, Entity]:
        for entity in self._entities:
            if entity_id == entity.id:
                return entity
        raise KeyError(f'Cannot find an entity with ID "{entity_id}".')

    @override
    def __delitem__(self, key: str | Intersection[TargetT, Entity]) -> None:
        if isinstance(key, str):
            for entity in self._entities:
                if entity.id == key:
                    self.remove(entity)
        else:
            self.remove(key)

    @override
    def __contains__(self, value: Any) -> bool:
        if isinstance(value, str):
            return any(entity.id == value for entity in self._entities)
        return any(entity is value for entity in self._entities)


class MultipleTypesEntityCollection[TargetT = Entity](EntityCollection[TargetT]):
    """
    Collect entities of multiple types.
    """

    def __init__(self, *entities: Intersection[TargetT, Entity]):
        super().__init__()
        self._collections: MutableMapping[str, SingleTypeEntityCollection] = (
            defaultdict(SingleTypeEntityCollection)
        )
        self.add(*entities)

    def _get_collection[EntityT: Entity](
        self, entity_type: type[EntityT] | EntityDefinition | ResolvableMachineName, /
    ) -> SingleTypeEntityCollection[EntityT]:
        if isinstance(entity_type, EntityDefinition):
            entity_type = entity_type.id
        if isinstance(entity_type, type):
            entity_type = entity_type.plugin().id
        return cast(SingleTypeEntityCollection[EntityT], self._collections[entity_type])

    def __getitem__[EntityT: Entity](
        self,
        key: type[EntityT] | EntityDefinition | ResolvableMachineName,
    ) -> SingleTypeEntityCollection[EntityT]:
        return self._get_collection(key)

    @override
    def __delitem__(self, key: Intersection[TargetT, Entity]) -> None:
        self.remove(key)

    @override
    def __iter__(self) -> Iterator[Intersection[TargetT, Entity]]:
        for collection in self._collections.values():
            for entity in collection:
                yield cast("Intersection[TargetT , Entity]", entity)

    @override
    def __len__(self) -> int:
        return sum(map(len, self._collections.values()))

    @override
    def __contains__(self, value: Any) -> bool:
        if isinstance(value, Entity):
            return any(entity is value for entity in self)
        return False

    @override
    def add(self, *entities: Intersection[TargetT, Entity]) -> None:
        added_entities = [*self._unknown(*entities)]
        for entity in added_entities:
            self[type(entity)].add(entity)
        if added_entities:
            self._on_add(*added_entities)

    @override
    def remove(self, *entities: Intersection[TargetT, Entity]) -> None:
        removed_entities = [*self._known(*entities)]
        for entity in removed_entities:
            self[type(entity)].remove(entity)
        if removed_entities:
            self._on_remove(*removed_entities)

    @override
    def clear(self) -> None:
        removed_entities = (*self,)
        for collection in self._collections.values():
            collection.clear()
        if removed_entities:
            self._on_remove(*removed_entities)


@contextmanager
def record_added[EntityT: Entity](
    entities: EntityCollection[EntityT], /
) -> Iterator[MultipleTypesEntityCollection[EntityT]]:
    """
    Record all entities that are added to a collection.
    """
    original = [*entities]
    added = MultipleTypesEntityCollection[EntityT]()
    yield added
    added.add(*[entity for entity in entities if entity not in original])



"""
Provide caching that stores cache items in volatile memory.
"""

from __future__ import annotations

from collections.abc import MutableMapping, Sequence
from typing import TYPE_CHECKING, Self, final, override

from betty.cache import CacheItem
from betty.cache._base import _CommonCacheBase, _CommonCacheBaseState, _StaticCacheItem
from betty.typing import threadsafe

if TYPE_CHECKING:
    from betty.concurrent import AsynchronizedLock, Ledger


type _MemoryCacheStore[CacheItemValueT] = MutableMapping[
    tuple[str, ...],
    "CacheItem[CacheItemValueT] | None | _MemoryCacheStore[CacheItemValueT]",
]


@final
class _MemoryCacheState[CacheItemValueT](
    _CommonCacheBaseState["MemoryCache[CacheItemValueT]"],
):
    def __init__(
        self,
        cache_lock: AsynchronizedLock,
        cache_item_lock_ledger: Ledger,
        store: _MemoryCacheStore[CacheItemValueT],
    ):
        super().__init__(cache_lock, cache_item_lock_ledger)
        self.store = store


@final
@threadsafe
class MemoryCache[CacheItemValueT](_CommonCacheBase[CacheItemValueT]):
    """
    Provide a cache that stores cache items in volatile memory.
    """

    _store: _MemoryCacheStore[CacheItemValueT]

    def __init__(
        self,
        *,
        scopes: Sequence[str] | None = None,
        state: _MemoryCacheState[CacheItemValueT] | None = None,
    ):
        super().__init__(scopes=scopes, state=state)
        if state is None:
            self._store = {}
        else:
            self._store = state.store

    @override
    def with_scope(self, scope: str, /) -> Self:
        return type(self)(
            scopes=(*self._scopes, scope),
            state=_MemoryCacheState[CacheItemValueT](
                self._cache_lock, self._cache_item_lock_ledger, self._store
            ),
        )

    def _cache_item_key(self, cache_item_id: str) -> tuple[str, ...]:
        return *self._scopes, cache_item_id

    @override
    async def get(self, cache_item_id: str, /) -> CacheItem[CacheItemValueT] | None:
        cache_item = self._store.get(self._cache_item_key(cache_item_id), None)
        if isinstance(cache_item, CacheItem):
            return cache_item  # ty:ignore[invalid-return-type]
        return None

    @override
    async def set(
        self,
        cache_item_id: str,
        value: CacheItemValueT,
        *,
        modified: int | float | None = None,
    ) -> None:
        self._store[self._cache_item_key(cache_item_id)] = _StaticCacheItem(
            value, modified
        )

    @override
    async def delete(self, cache_item_id: str, /) -> None:
        self._store.pop(self._cache_item_key(cache_item_id), None)

    @override
    async def clear(self) -> None:
        self._store.clear()



"""
Provide error handling utilities.
"""

from pathlib import Path

from betty.exception import HumanFacingException
from betty.locale.localizable.gettext import _


class FileNotFound(HumanFacingException, FileNotFoundError):
    """
    Raised when a file cannot be found.
    """

    def __init__(self, file_path: Path, /):
        super().__init__(
            _('Could not find the file "{file_path}".').format(file_path=str(file_path))
        )



"""
Provide the Ancestry loading API.
"""

from abc import ABC, abstractmethod
from asyncio import gather

from betty.ancestry.link import Link
from betty.concurrent import MAX_STRANDS
from betty.job import Context
from betty.job.executor.asyncio import AsyncExecutor
from betty.job.scheduler import Scheduler
from betty.job.scheduler.default import DefaultScheduler
from betty.project import Project
from betty.project.load.jobs import PopulateLink


class Loader(ABC):
    """
    Load ancestry data into a project.
    """

    @abstractmethod
    async def load(self, scheduler: Scheduler) -> None:
        """
        Load ancestry data into a project.
        """


class PostLoader(ABC):
    """
    Postprocess ancestry data after it has been loaded.
    """

    @abstractmethod
    async def post_load(self, scheduler: Scheduler) -> None:
        """
        Postprocess ancestry data after it has been loaded.
        """


async def load(project: Project, *, context: Context | None = None) -> None:
    """
    Load an ancestry.
    """
    if context is None:
        context = Context()

    app = project.app
    http_client = await app.http_client
    localizers = await project.public_localizers

    extensions = await project.extensions
    load_scheduler = DefaultScheduler(context=context, user=project.app.user)
    async with AsyncExecutor(load_scheduler, concurrency=MAX_STRANDS):
        await gather(
            *(
                extension.load(load_scheduler)
                for extension in extensions
                if isinstance(extension, Loader)
            )
        )
        await load_scheduler.release()
        await load_scheduler.complete()
    post_load_scheduler = DefaultScheduler(context=context, user=project.app.user)
    async with AsyncExecutor(post_load_scheduler, concurrency=MAX_STRANDS):
        await gather(
            *(
                extension.post_load(post_load_scheduler)
                for extension in extensions
                if isinstance(extension, PostLoader)
            )
        )
        await post_load_scheduler.release()
        await post_load_scheduler.add(
            *(
                PopulateLink(link, http_client=http_client, localizers=localizers)
                for link in project.ancestry[Link]
            ),
        )
        await post_load_scheduler.complete()



"""
Jobs for the Maps extension.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, override

from betty.ancestry.place import Place
from betty.job import Job
from betty.project.generate.file import create_file

if TYPE_CHECKING:
    from babel import Locale

    from betty.job.scheduler import Scheduler
    from betty.project import Project


class _GeneratePlacePreviews(Job):
    def __init__(self, *, project: Project):
        super().__init__("maps:generate-place-previews", priority=True)
        self._project = project

    @override
    async def do(self, scheduler: Scheduler, /) -> None:
        await scheduler.add(
            *(
                _GeneratePlacePreview(place.id, locale, self._project)
                for locale in self._project.configuration.locales.keys()  # noqa: SIM118
                for place in self._project.ancestry[Place]
                if place.coordinates
            )
        )


class _GeneratePlacePreview(Job):
    def __init__(self, place_id: str, locale: Locale, project: Project):
        super().__init__(f"maps:generate-place-preview:{place_id}:{locale}")
        self._place_id = place_id
        self._locale = locale
        self._project = project

    @override
    async def do(self, scheduler: Scheduler, /) -> None:
        context = scheduler.context
        place = self._project.ancestry[Place][self._place_id]
        app = self._project.app
        localizers = await app.localizers
        jinja = await self._project.jinja
        place_path = (
            self._project.localize_www_directory(self._locale)
            / place.plugin().id
            / place.public_id
        )
        rendered_html = await jinja.get_template(
            "component/maps/selected-place-preview.html.j2",
        ).render_async(
            document=await self._project.new_document(
                context=context, localizer=localizers.get(self._locale)
            ),
            place=place,
        )
        async with create_file(place_path / "-maps-place-preview.html") as f:
            await f.write(rendered_html)



"""
Test utilities for :py:mod:`betty.plugin`.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final, override

from betty.plugin import Plugin, PluginDefinition, PluginTypeDefinition
from betty.plugin.factory import PluginManufacturer
from betty.test_utils.locale.localizable import DUMMY_COUNTABLE_LOCALIZABLE

if TYPE_CHECKING:
    import builtins


class DummyPlugin(Plugin["DummyPluginDefinition"]):
    """
    A dummy plugin.
    """


@final
@PluginTypeDefinition(
    "dummy-plugin",
    label=" dummy plugin",
    label_plural=" dummy plugin",
    label_countable=DUMMY_COUNTABLE_LOCALIZABLE,
)
class DummyPluginDefinition(PluginDefinition[DummyPlugin]):
    """
    A definition of a dummy plugin.
    """


@final
@DummyPluginDefinition("dummy-plugin-one")
class DummyPluginOne(DummyPlugin):
    """
    A dummy plugin (one).
    """


@final
@DummyPluginDefinition("dummy-plugin-two")
class DummyPluginTwo(DummyPlugin):
    """
    A dummy plugin (two).
    """


@final
@DummyPluginDefinition("dummy-plugin-three")
class DummyPluginThree(DummyPlugin):
    """
    A dummy plugin (three).
    """


@final
@DummyPluginDefinition("dummy-plugin-four")
class DummyPluginFour(DummyPlugin):
    """
    A dummy plugin (four).
    """


@final
class DummyPluginManufacturer(PluginManufacturer[DummyPluginDefinition, DummyPlugin]):
    """
    The dummy plugin manufacturer.
    """

    @override
    @classmethod
    def type(cls) -> builtins.type[DummyPluginDefinition]:
        return DummyPluginDefinition



"""
Provide rendering utilities using `Jinja2 <https://jinja.palletsprojects.com>`_.
"""

from __future__ import annotations

import datetime
from collections.abc import Awaitable, Callable, Iterable, Mapping, MutableMapping
from pathlib import Path
from shutil import copy2
from typing import TYPE_CHECKING, Any, Self, cast, override

import aiofiles
from aiofiles.os import makedirs
from jinja2 import Environment as JinjaEnvironment
from jinja2 import FileSystemLoader, pass_context, select_autoescape
from jinja2.async_utils import auto_await
from jinja2.ext import Extension as JinjaExtension
from jinja2.nodes import CallBlock, ContextReference, Node
from jinja2.runtime import Context as JinjaContext
from jinja2.runtime import DebugUndefined, StrictUndefined
from jinja2.utils import missing

from betty import about
from betty.cache import CacheItem
from betty.date import Date
from betty.html import CssProvider, JsProvider, NavigationLinkProvider, generate_html_id
from betty.html.attributes import Attributes
from betty.jinja.filter import filters
from betty.jinja.test import tests
from betty.media_type import UnsupportedMediaType, match_extension
from betty.media_type.media_types import JINJA2
from betty.service.factory import Manufacturable
from betty.service.requirement.project import require_project
from betty.warnings import deprecate

if TYPE_CHECKING:
    from jinja2.parser import Parser

    from betty.asset import AssetRepository
    from betty.document import Document
    from betty.extension import Extension
    from betty.job import Context as JobContext
    from betty.locale.localize import Localizer
    from betty.project import Project


type CopyFunction = Callable[[Path, Path], Awaitable[None]]


def context_project(context: JinjaContext) -> Project:
    """
    Get the current project from the Jinja2 context.
    """
    return cast(Environment, context.environment).project


def context_document(context: JinjaContext) -> Document:
    """
    Get the current document from the Jinja2 context.
    """
    document: Document = context.resolve_or_missing("document")
    if document is missing:
        raise RuntimeError(
            "No `document` context variable exists in this Jinja2 template."
        ) from None
    return document


def context_context(context: JinjaContext) -> JobContext | None:
    """
    Get the current job context from the Jinja2 context.
    """
    try:
        return context_document(context).context
    except (KeyError, RuntimeError):
        return None


def context_localizer(context: JinjaContext) -> Localizer:
    """
    Get the current localizer from the Jinja2 context.
    """
    try:
        return context_document(context).localizer
    except KeyError:
        raise RuntimeError(
            "No `resource.localizer` context variable exists in this Jinja2 template."
        ) from None


type Globals = Mapping[str, Any]
type Filters = Mapping[str, Callable[..., Any]]
type Tests = Mapping[str, Callable[..., bool]]


class JinjaProvider:
    """
    Integrate an :py:class:`betty.extension.Extension` with the Jinja2 API.
    """

    @property
    def globals(self) -> Globals:
        """
        Jinja2 globals provided by this extension.

        Keys are the globals' names, and values are the globals' values.
        """
        return {}

    @property
    def filters(self) -> Filters:
        """
        Jinja2 filters provided by this extension.

        Keys are filter names, and values are the filters themselves.
        """
        return {}

    @property
    def tests(self) -> Tests:
        """
        Jinja2 tests provided by this extension.

        Keys are test names, and values are the tests themselves.
        """
        return {}


class Environment(Manufacturable, JinjaEnvironment):
    """
    Betty's Jinja2 environment.
    """

    globals: dict[str, Any]
    filters: dict[str, Callable[..., Any]]
    tests: dict[str, Callable[..., bool]]

    def __init__(
        self,
        project: Project,
        extensions: Iterable[Extension],
        assets: AssetRepository,
        globals: Mapping[str, Any],  # noqa: A002
        filters: Mapping[str, Callable[..., Any]],
        tests: Mapping[str, Callable[..., bool]],
    ):
        template_directory_paths = [
            str(path / "templates") for path in assets.directories
        ]
        super().__init__(
            loader=FileSystemLoader(template_directory_paths),
            auto_reload=project.configuration.debug,
            enable_async=True,
            undefined=(
                DebugUndefined if project.configuration.debug else StrictUndefined
            ),
            autoescape=select_autoescape(["html.j2"]),
            trim_blocks=True,
            lstrip_blocks=True,
            extensions=[
                "jinja2.ext.do",
                "jinja2.ext.i18n",
                _CacheTagExtension,
            ],
        )

        self._project = project
        self._extensions = tuple(extensions)

        if project.configuration.debug:
            self.add_extension("jinja2.ext.debug")

        self._init_i18n()
        self._init_globals()
        self.globals.update(globals)
        self.filters.update(filters)
        self.tests.update(tests)
        self._init_extensions()

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        extensions = await project.extensions
        return cls(
            project,
            extensions,
            await project.assets,
            {
                # Ideally we would use the Dispatcher for this. However, it is asynchronous only.
                "public_css_paths": [
                    path
                    for extension in extensions
                    if isinstance(extension, CssProvider)
                    for path in await extension.get_public_css_paths()
                ],
                "public_js_paths": [
                    path
                    for extension in extensions
                    if isinstance(extension, JsProvider)
                    for path in await extension.get_public_js_paths()
                ],
            },
            await filters(),
            await tests(),
        )

    @property
    def project(self) -> Project:
        """
        The current project.
        """
        return self._project

    def _init_i18n(self) -> None:
        self.install_gettext_callables(  # ty:ignore[unresolved-attribute]
            gettext=self._gettext,
            ngettext=self._ngettext,
            pgettext=self._pgettext,
            npgettext=self._npgettext,
            newstyle=True,
        )
        self.policies["ext.i18n.trimmed"] = True

    @pass_context
    def _gettext(self, context: JinjaContext, message: str) -> str:
        return context_localizer(context).gettext(message)

    @pass_context
    def _ngettext(
        self, context: JinjaContext, message_singular: str, message_plural: str, n: int
    ) -> str:
        return context_localizer(context).ngettext(message_singular, message_plural, n)

    @pass_context
    def _pgettext(
        self, context: JinjaContext, gettext_context: str, message: str
    ) -> str:
        return context_localizer(context).pgettext(gettext_context, message)

    @pass_context
    def _npgettext(
        self,
        context: JinjaContext,
        gettext_context: str,
        message_singular: str,
        message_plural: str,
        n: int,
    ) -> str:
        return context_localizer(context).npgettext(
            gettext_context, message_singular, message_plural, n
        )

    def _init_globals(self) -> None:
        self.globals["about_version_major"] = about.VERSION_MAJOR_LABEL
        self.globals["app"] = self.project.app
        self.globals["project"] = self.project
        today = datetime.date.today()
        self.globals["today"] = Date(today.year, today.month, today.day)
        self.globals["primary_navigation_links"] = [
            link
            for extension in self._extensions
            if isinstance(extension, NavigationLinkProvider)
            for link in extension.primary_navigation_links()
        ]
        self.globals["secondary_navigation_links"] = [
            link
            for extension in self._extensions
            if isinstance(extension, NavigationLinkProvider)
            for link in extension.secondary_navigation_links()
        ]
        self.globals["generate_html_id"] = generate_html_id
        self.globals["deprecate"] = deprecate
        self.globals["new_attributes"] = Attributes

    def _init_extensions(self) -> None:
        for extension in self._extensions:
            if isinstance(extension, JinjaProvider):
                self.globals.update(extension.globals)
                self.filters.update(extension.filters)
                self.tests.update(extension.tests)

    def make_copy_function(
        self,
        *,
        document: Document,
        www_directory_path: Path | None = None,
        is_localized_and_multilingual: bool | None = None,
    ) -> CopyFunction:
        """
        Make a copy function for this renderer that renders supported files.
        """

        async def _copy_function(source_path: Path, destination_path: Path) -> None:
            await makedirs(destination_path.parent, exist_ok=True)
            try:
                media_type, extension = match_extension(source_path, [JINJA2])
            except UnsupportedMediaType:
                copy2(source_path, destination_path)
                return

            destination_path = destination_path.with_name(
                destination_path.name[: -len(extension)]
            )

            copy_resource_url = document.resource_url

            if www_directory_path:
                try:
                    relative_file_destination_path = destination_path.relative_to(
                        www_directory_path
                    )
                except ValueError:
                    pass
                else:
                    resource_parts = relative_file_destination_path.parts
                    if not any(
                        resource_part.startswith(".")
                        for resource_part in resource_parts
                    ):
                        if is_localized_and_multilingual:
                            resource_parts = resource_parts[1:]
                        copy_resource_url = f"betty:///{'/'.join(resource_parts)}"
            async with aiofiles.open(source_path) as f:
                content = await f.read()

            template = self.from_string(content)
            copy_document = document.copy(
                resource=destination_path, resource_url=copy_resource_url
            )
            rendered_content = await template.render_async(document=copy_document)
            async with aiofiles.open(destination_path, "w") as f:
                await f.write(rendered_content)

        return _copy_function


type _CacheExtensionMap = MutableMapping[str, str]


class _CacheTagExtension(JinjaExtension):
    tags = {"cache"}

    @override
    def parse(self, parser: Parser) -> Node | list[Node]:
        lineno = next(parser.stream).lineno
        cache_key = parser.parse_expression()
        body = parser.parse_statements(("name:endcache",), drop_needle=True)
        return CallBlock(
            self.call_method("_cache", [cache_key, ContextReference()]),
            [],
            [],
            body,
        ).set_lineno(lineno)

    async def _cache(
        self, cache_key: str, context: JinjaContext, caller: Callable[[], str]
    ) -> str:
        job_context = context_context(context)
        if job_context is None:
            return await auto_await(caller())
        async with job_context.cache.getset(f"jinja2_cache_tag:{cache_key}") as result:
            if isinstance(result, CacheItem):
                return cast(str, await result.value())
            rendered = await auto_await(caller())
            await result(rendered)
            return rendered



"""
Provide caching that persists cache items to files.
"""

from __future__ import annotations

import asyncio
import shutil
from abc import abstractmethod
from asyncio import to_thread
from contextlib import asynccontextmanager, suppress
from functools import partial
from os import utime
from pickle import dumps, loads
from typing import TYPE_CHECKING, Self, final, override

import aiofiles
from aiofiles.os import makedirs, remove
from aiofiles.ospath import getmtime

from betty.cache import CacheItem, CacheItemValueSetter
from betty.cache._base import _CommonCacheBase, _CommonCacheBaseState
from betty.hashid import hashid
from betty.typing import threadsafe

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Sequence
    from pathlib import Path


class _FileCacheItem[CacheItemValueT](CacheItem[CacheItemValueT]):
    __slots__ = "_modified", "_path"

    def __init__(
        self,
        modified: int | float,
        path: Path,
    ):
        self._modified = modified
        self._path = path

    @override
    @property
    def modified(self) -> int | float:
        return self._modified

    @override
    async def value(self) -> CacheItemValueT:
        async with aiofiles.open(self._path, "rb") as f:
            value_bytes = await f.read()
        return await self._load_value(value_bytes)

    @abstractmethod
    async def _load_value(self, value_bytes: bytes) -> CacheItemValueT:
        pass


@final
class _PickledFileCacheItem[CacheItemValueT](_FileCacheItem[CacheItemValueT]):
    @override
    async def _load_value(self, value_bytes: bytes) -> CacheItemValueT:
        return loads(value_bytes)


@final
class _BinaryFileCacheItem(_FileCacheItem[bytes]):
    @override
    async def _load_value(self, value_bytes: bytes) -> bytes:
        return value_bytes


class _FileCache[CacheItemValueT](_CommonCacheBase[CacheItemValueT]):
    """
    Provide a cache that persists cache items on a file system.
    """

    _cache_item_cls: type[_FileCacheItem[CacheItemValueT]]

    def __init__(
        self,
        cache_directory_path: Path,
        *,
        scopes: Sequence[str] | None = None,
        state: _CommonCacheBaseState[Self] | None = None,
    ):
        super().__init__(scopes=scopes, state=state)
        self._root_path = cache_directory_path

    @override
    def with_scope(self, scope: str, /) -> Self:
        return type(self)(
            self._root_path,
            scopes=(*self._scopes, scope),
            state=_CommonCacheBaseState(self._cache_lock, self._cache_item_lock_ledger),
        )

    def _cache_item_file_path(
        self, cache_item_id: str, suffix: str | None = None
    ) -> Path:
        cache_item_file_path = self._path / hashid(cache_item_id)
        if suffix is not None:
            assert suffix.startswith(".")
            cache_item_file_path = cache_item_file_path.parent / (
                cache_item_file_path.name + suffix
            )
        return cache_item_file_path

    @abstractmethod
    def _dump_value(self, value: CacheItemValueT) -> bytes:
        pass

    @override
    async def has(self, cache_item_id: str, *, suffix: str | None = None) -> bool:
        return await to_thread(self._cache_item_file_path(cache_item_id, suffix).exists)

    @override
    async def get(
        self, cache_item_id: str, *, suffix: str | None = None
    ) -> CacheItem[CacheItemValueT] | None:
        try:
            cache_item_file_path = self._cache_item_file_path(cache_item_id, suffix)
            return self._cache_item_cls(
                await getmtime(cache_item_file_path),
                cache_item_file_path,
            )
        except OSError:
            return None

    @override
    async def set(
        self,
        cache_item_id: str,
        value: CacheItemValueT,
        *,
        suffix: str | None = None,
        modified: int | float | None = None,
    ) -> None:
        value_bytes = self._dump_value(value)
        cache_item_file_path = self._cache_item_file_path(cache_item_id, suffix)
        try:
            await self._write(cache_item_file_path, value_bytes, modified)
        except FileNotFoundError:
            await makedirs(cache_item_file_path.parent, exist_ok=True)
            await self._write(cache_item_file_path, value_bytes, modified)

    async def _write(
        self,
        cache_item_file_path: Path,
        value: bytes,
        modified: int | float | None = None,
    ) -> None:
        async with aiofiles.open(cache_item_file_path, "wb") as f:
            await f.write(value)
        if modified is not None:
            await asyncio.to_thread(utime, cache_item_file_path, (modified, modified))

    @override
    async def delete(self, cache_item_id: str, *, suffix: str | None = None) -> None:
        with suppress(FileNotFoundError):
            await remove(self._cache_item_file_path(cache_item_id, suffix))

    @override
    async def clear(self) -> None:
        with suppress(FileNotFoundError):
            await asyncio.to_thread(shutil.rmtree, self._path)

    @property
    def _path(self) -> Path:
        return self._root_path.joinpath(*self._scopes)

    @override
    @asynccontextmanager
    async def hasset(
        self, cache_item_id: str, *, suffix: str | None = None
    ) -> AsyncIterator[CacheItemValueSetter[CacheItemValueT] | None]:
        if await self.has(cache_item_id, suffix=suffix):
            yield None
            return
        async with self._cache_item_lock_ledger.ledger(cache_item_id):
            if await self.has(cache_item_id, suffix=suffix):
                yield None
                return
            yield partial(self.set, cache_item_id, suffix=suffix)
        return

    @override
    @asynccontextmanager
    async def getset(
        self, cache_item_id: str, *, suffix: str | None = None
    ) -> AsyncIterator[
        CacheItemValueSetter[CacheItemValueT] | CacheItem[CacheItemValueT]
    ]:
        if cache_item := await self.get(cache_item_id):
            yield cache_item
            return
        async with self._cache_item_lock_ledger.ledger(cache_item_id):
            if cache_item := await self.get(cache_item_id, suffix=suffix):
                yield cache_item
                return
            yield partial(self.set, cache_item_id, suffix=suffix)
        return


@final
@threadsafe
class PickledFileCache[CacheItemValueT](_FileCache[CacheItemValueT]):
    """
    Provide a cache that pickles values and persists them to files.
    """

    _cache_item_cls = _PickledFileCacheItem

    @override
    def _dump_value(self, value: CacheItemValueT) -> bytes:
        return dumps(value)


@final
@threadsafe
class BinaryFileCache(_FileCache[bytes]):
    """
    Provide a cache that persists bytes values to binary files.
    """

    _cache_item_cls = _BinaryFileCacheItem

    @override
    def _dump_value(self, value: bytes) -> bytes:
        return value

    @property
    def path(self) -> Path:
        """
        The path to the cache's root directory.
        """
        return self._path

    def cache_item_file_path(
        self, cache_item_id: str, suffix: str | None = None, /
    ) -> Path:
        """
        Get the file path for a cache item with the given ID.

        The cache item itself may or may not exist.
        """
        return self._cache_item_file_path(cache_item_id, suffix)



"""
The Assertion API.
"""

from __future__ import annotations

from collections.abc import (
    Callable,
    Mapping,
    MutableMapping,
    MutableSequence,
    Sequence,
    Sized,
)
from dataclasses import dataclass
from enum import Enum
from functools import lru_cache
from pathlib import Path
from types import NoneType
from typing import TYPE_CHECKING, Any, final, overload

from betty.data.indicator.selector import Index, Key
from betty.error import FileNotFound
from betty.exception import HumanFacingException, reraise_with_indicator
from betty.locale import from_language_tag
from betty.locale.localizable.gettext import _
from betty.locale.localizable.markup import Paragraph, Paragraphs, do_you_mean

if TYPE_CHECKING:
    from babel import Locale

    from betty.locale.localizable import Localizable

type Number = int | float

type Assertion[ValueT, ReturnT] = Callable[[ValueT], ReturnT]


class AssertionChain[ValueT, ReturnT]:
    """
    An assertion chain.

    Assertion chains let you chain/link/combine assertions into pipelines that take an input
    value and, if the assertions pass, return an output value. Each chain may be (re)used as many
    times as needed.

    Assertion chains are assertions themselves: you can use a chain wherever you can use a 'plain'
    assertion.

    Assertions chains are `monads <https://en.wikipedia.org/wiki/Monad_(functional_programming)>`_.
    While uncommon in Python, this allows us to create these chains in a type-safe way, and type checkers
    can confirm that all assertions in any given chain are compatible with each other.
    """

    def __init__(self, _assertion: Assertion[ValueT, ReturnT], /):
        self._assertion = _assertion

    def chain[AssertionsExtendReturnT](
        self, assertion: Assertion[ReturnT, AssertionsExtendReturnT], /
    ) -> AssertionChain[ValueT, AssertionsExtendReturnT]:
        """
        Extend the chain with the given assertion.
        """
        return AssertionChain(lambda value: assertion(self(value)))

    def __or__[AssertionsExtendReturnT](
        self, _assertion: Assertion[ReturnT, AssertionsExtendReturnT]
    ) -> AssertionChain[ValueT, AssertionsExtendReturnT]:
        return self.chain(_assertion)

    def __call__(self, value: ValueT) -> ReturnT:
        """
        Invoke the chain with a value.

        This method may be called more than once.

        :raises betty.exception.HumanFacingException: Raised if any part of the
            assertion chain fails.
        """
        return self._assertion(value)


@dataclass(frozen=True)
class Field[ValueT, ReturnT]:
    """
    A key-value mapping field.

    Do not instantiate this class directly. Use :py:class:`betty.assertion.RequiredField` or
    :py:class:`betty.assertion.OptionalField` instead.
    """

    name: str
    assertion: Assertion[ValueT, ReturnT] | None = None
    as_name: str | None = None


@final
@dataclass(frozen=True)
class RequiredField[ValueT, ReturnT](Field[ValueT, ReturnT]):
    """
    A required key-value mapping field.
    """


@final
@dataclass(frozen=True)
class OptionalField[ValueT, ReturnT](Field[ValueT, ReturnT]):
    """
    An optional key-value mapping field.
    """


type _AssertionBuilderFunction[ValueT, ReturnT] = Callable[[ValueT], ReturnT]
type _AssertionBuilderMethod[ValueT, ReturnT] = Callable[[object, ValueT], ReturnT]
type _AssertionBuilder[ValueT, ReturnT] = (
    "_AssertionBuilderFunction[ValueT, ReturnT] | _AssertionBuilderMethod[ValueT, ReturnT]"
)


type AssertTypeType = (
    bool | float | int | Mapping[Any, Any] | None | Sequence[Any] | str
)


_ASSERT_TYPES: Mapping[type[AssertTypeType], tuple[type[Any] | None, Localizable]] = {
    bool: (None, _("This must be a boolean.")),
    int: (bool, _("This must be a whole number.")),
    float: (None, _("This must be a decimal number.")),
    Mapping: (None, _("This must be a key-value mapping.")),
    NoneType: (None, _("This must be none/null.")),
    Sequence: (None, _("This must be a sequence.")),
    str: (None, _("This must be a string.")),
}


@lru_cache
def assert_type[AssertTypeTypeT: AssertTypeType](
    value_type: type[AssertTypeTypeT], /
) -> AssertionChain[Any, AssertTypeTypeT]:
    """
    Assert that a value is of the specified built-in type.
    """

    def _assert_type(value: Any, /) -> AssertTypeTypeT:
        value_is_not_type, error_message = _ASSERT_TYPES[value_type]
        if isinstance(value, value_type) and (
            value_is_not_type is None or not isinstance(value, value_is_not_type)
        ):
            return value
        raise HumanFacingException(error_message)

    return AssertionChain(_assert_type)


def assert_or[ValueT, ReturnT, AssertionReturnU](
    if_assertion: Assertion[ValueT, ReturnT],
    else_assertion: Assertion[ValueT, AssertionReturnU],
    /,
) -> AssertionChain[ValueT, ReturnT | AssertionReturnU]:
    """
    Assert that at least one of the given assertions passed.
    """

    def _assert_or(value: Any, /) -> ReturnT | AssertionReturnU:
        assertions = (if_assertion, else_assertion)
        errors = []
        for assertion in assertions:
            try:
                return assertion(value)
            except HumanFacingException as e:
                errors.append(e)
        raise HumanFacingException(Paragraphs(*errors))

    return AssertionChain(_assert_or)


assert_none = assert_type(NoneType)
"""
Assert that a value is ``None``.
"""


assert_bool = assert_type(bool)
"""
Assert that a value is a Python ``bool``.
"""


def _assert_number[NumberT: Number](
    minimum: Number | None = None, maximum: Number | None = None
) -> AssertionChain[NumberT, NumberT]:
    def __assert_number(value: NumberT) -> NumberT:
        if minimum is not None and value < minimum:
            raise HumanFacingException(
                _("This must be at least {minimum}.").format(minimum=str(minimum))
            )
        if maximum is not None and value > maximum:
            raise HumanFacingException(
                _("This must be at most {maximum}.").format(maximum=str(maximum))
            )
        return value

    return AssertionChain(__assert_number)


def assert_int(
    *, minimum: Number | None = None, maximum: Number | None = None
) -> AssertionChain[Any, int]:
    """
    Assert that a value is a Python ``int``.
    """
    return assert_type(int) | _assert_number(minimum, maximum)


def assert_float(
    *, minimum: Number | None = None, maximum: Number | None = None
) -> AssertionChain[Any, float]:
    """
    Assert that a value is a Python ``float``.
    """
    return assert_type(float) | _assert_number(minimum, maximum)


def assert_number(
    *, minimum: Number | None = None, maximum: Number | None = None
) -> AssertionChain[Any, Number]:
    """
    Assert that a value is a number (a Python ``int`` or ``float``).
    """
    return assert_or(assert_int(), assert_float()) | _assert_number(minimum, maximum)


def assert_str(
    *,
    exact_length: int | None = None,
    minimum_length: int | None = None,
    maximum_length: int | None = None,
) -> AssertionChain[Any, str]:
    """
    Assert that a value is a Python ``str``.
    """

    def _assert_str(value: Any, /) -> str:
        string = assert_type(str)(value)
        actual = len(value)
        if exact_length is not None and actual != exact_length:
            raise HumanFacingException(
                _("This must be {length} characters long.").format(
                    length=str(exact_length)
                )
            )
        if minimum_length is not None and actual < minimum_length:
            raise HumanFacingException(
                _("This must be at least {length} characters long.").format(
                    length=str(minimum_length)
                )
            )
        if maximum_length is not None and actual > maximum_length:
            raise HumanFacingException(
                _("This must be at most {length} characters long.").format(
                    length=str(maximum_length)
                )
            )
        return string

    return AssertionChain(_assert_str)


@overload
def assert_sequence(
    value_assertion: None = None, /
) -> AssertionChain[Any, MutableSequence[Any]]:
    pass


@overload
def assert_sequence[ReturnT](
    value_assertion: Assertion[Any, ReturnT], /
) -> AssertionChain[Any, MutableSequence[ReturnT]]:
    pass


def assert_sequence[ReturnT](value_assertion: Assertion[Any, ReturnT] | None = None, /):
    """
    Assert that a value is a sequence.

    Optionally assert that values are of a given type.
    """

    def _assert_sequence(value: Any, /) -> MutableSequence[ReturnT]:
        sequence = assert_type(Sequence)(value)
        if value_assertion is None:
            return list(sequence)
        asserted_sequence = []
        for value_index, value_value in enumerate(sequence):
            with reraise_with_indicator(Index(value_index)):
                asserted_sequence.append(value_assertion(value_value))
        return asserted_sequence

    return AssertionChain(_assert_sequence)


@overload
def assert_mapping(
    value_assertion: None = None, key_assertion: None = None, /
) -> AssertionChain[Any, MutableMapping[Any, Any]]:
    pass


@overload
def assert_mapping[ReturnT](
    value_assertion: Assertion[Any, ReturnT], key_assertion: None = None, /
) -> AssertionChain[Any, MutableMapping[Any, ReturnT]]:
    pass


@overload
def assert_mapping[AssertionKeyT](
    value_assertion: None, key_assertion: Assertion[Any, AssertionKeyT], /
) -> AssertionChain[Any, MutableMapping[AssertionKeyT, Any]]:
    pass


@overload
def assert_mapping[ReturnT, AssertionKeyT](
    value_assertion: Assertion[Any, ReturnT],
    key_assertion: Assertion[Any, AssertionKeyT],
    /,
) -> AssertionChain[Any, MutableMapping[AssertionKeyT, ReturnT]]:
    pass


def assert_mapping[ReturnT, AssertionKeyT](
    value_assertion: Assertion[Any, ReturnT] | None = None,
    key_assertion: Assertion[Any, AssertionKeyT] | None = None,
    /,
):
    """
    Assert that a value is a key-value mapping.

    Optionally assert that keys and/or values are of a given type.
    """

    def _assert_mapping(value: Any, /) -> MutableMapping[AssertionKeyT, ReturnT]:
        mapping = assert_type(Mapping)(value)
        if value_assertion is None and key_assertion is None:
            return dict(mapping)
        asserted_mapping = {}
        for value_key, value_value in mapping.items():
            asserted_value_key = value_key
            if key_assertion:
                with reraise_with_indicator(Key(str(value_key))):
                    asserted_value_key = key_assertion(value_key)
            asserted_value_value = value_value
            if value_assertion:
                with reraise_with_indicator(Key(str(value_key))):
                    asserted_value_value = value_assertion(value_value)
            asserted_mapping[asserted_value_key] = asserted_value_value
        return asserted_mapping

    return AssertionChain(_assert_mapping)


def assert_record(
    *fields: Field[Any, Any], allow_extra: bool = False
) -> AssertionChain[Any, MutableMapping[str, Any]]:
    """
    Assert that a value is a record: a key-value mapping of arbitrary value types, with a known structure.

    To validate a key-value mapping as a records, assertions for all possible keys
    MUST be provided. Any keys present in the value for which no field assertions
    are provided will cause the entire record assertion to fail.
    """

    def _assert_record(value: Mapping[Any, Any], /) -> MutableMapping[str, Any]:
        known_keys = {x.name for x in fields}
        unknown_keys = set(value.keys()) - known_keys
        record: MutableMapping[str, Any] = {}
        if not allow_extra:
            for unknown_key in unknown_keys:
                with reraise_with_indicator(Key(unknown_key)):
                    raise HumanFacingException(
                        Paragraph(
                            _("Unknown key: {unknown_key}.").format(
                                unknown_key=f'"{unknown_key}"'
                            ),
                            do_you_mean(*(f'"{x}"' for x in sorted(known_keys))),
                        )
                    )
        for field in fields:
            with reraise_with_indicator(Key(field.name)):
                if field.name in value:
                    record[field.name if field.as_name is None else field.as_name] = (
                        field.assertion(value[field.name])
                        if field.assertion
                        else value[field.name]
                    )
                elif isinstance(field, RequiredField):
                    raise HumanFacingException(_("This field is required."))
        return record

    return assert_mapping() | _assert_record


def assert_isinstance[ValueT](alleged_type: type[ValueT], /) -> Assertion[Any, ValueT]:
    """
    Assert that a value is an instance of the given type.

    This assertion is **NOT** optimized to be user-facing (it is untranslated)
    because Python types are not user-facing.
    """

    def _assert(value: Any, /) -> ValueT:
        if isinstance(value, alleged_type):
            return value
        raise HumanFacingException(f"{value} must be an instance of {alleged_type}.")

    return _assert


def assert_path() -> AssertionChain[Any, Path]:
    """
    Assert that a value is a path to a file or directory on disk that may or may not exist.
    """
    return assert_or(assert_isinstance(Path), assert_str() | Path)


def assert_directory_path() -> AssertionChain[Any, Path]:
    """
    Assert that a value is a path to an existing directory.
    """

    def _assert_directory_path(directory_path: Path, /) -> Path:
        if directory_path.is_dir():
            return directory_path
        raise HumanFacingException(
            _('"{path}" is not a directory.').format(path=str(directory_path))
        )

    return assert_path() | _assert_directory_path


def assert_file_path() -> AssertionChain[Any, Path]:
    """
    Assert that a value is a path to an existing file.
    """

    def _assert_file_path(file_path: Path, /) -> Path:
        if file_path.is_file():
            return file_path
        raise FileNotFound(file_path)

    return assert_path() | _assert_file_path


def assert_locale() -> AssertionChain[Any, Locale]:
    """
    Assert that a value is a valid `IETF BCP 47 language tag <https://en.wikipedia.org/wiki/IETF_language_tag>`_.
    """
    return assert_str() | from_language_tag


def assert_len[SizedT: Sized](
    exact: int | None = None, *, minimum: int | None = None, maximum: int | None = None
) -> AssertionChain[SizedT, SizedT]:
    """
    Assert the length of a value.

    This assertion can be used in two ways:
    - with an exact required length
    - with minimum and/or maximum bounds (inclusive)
    """

    def _assert_len(value: SizedT, /) -> SizedT:
        actual = len(value)
        if exact is not None and actual != exact:
            raise HumanFacingException(
                _("Exactly {expected} items are required, but found {actual}.").format(
                    expected=str(exact), actual=str(actual)
                )
            )
        if minimum is not None and actual < minimum:
            raise HumanFacingException(
                _("At least {expected} items are required, but found {actual}.").format(
                    expected=str(minimum), actual=str(actual)
                )
            )
        if maximum is not None and actual > maximum:
            raise HumanFacingException(
                _("At most {expected} items are allowed, but found {actual}.").format(
                    expected=str(maximum), actual=str(actual)
                )
            )
        return value

    return AssertionChain(_assert_len)


def assert_enum[EnumT: Enum](options: type[EnumT]) -> AssertionChain[Any, EnumT]:
    """
    Assert that a value is allowed by an enum, and return the enum value.
    """

    def _assert_enum(value: Any) -> Any:
        try:
            return options(value)
        except ValueError:
            raise HumanFacingException(
                Paragraph(
                    _("Invalid option {value}.").format(value=str(value)),
                    do_you_mean(*[option.value for option in options]),
                )
            ) from None

    return AssertionChain(_assert_enum)



"""
JSON schemas for the privacy API.
"""

from __future__ import annotations

from typing import final

from betty.classtools import Singleton
from betty.json.schema import Boolean


@final
class PrivacySchema(Singleton, Boolean):
    """
    A JSON Schema for privacy.
    """

    def __init__(self):
        super().__init__(
            def_name="privacy",
            title="Privacy",
            description="Whether this entity is private (true), or public (false).",
        )



"""
Test utilities for :py:mod:`betty.serde`.
"""

import pytest

from betty.portable import PortableData
from betty.serde import Serializer


class SerializerTestBase[SerializerT: Serializer]:
    """
    A base class for testing :py:class:`betty.serde.Serializer` implementations.
    """

    @pytest.fixture
    def sut(self) -> type[SerializerT]:
        """
        Provide the system(s) under test.
        """
        raise NotImplementedError

    @pytest.mark.parametrize(
        "portable",
        [
            True,
            False,
            None,
            "abc",
            123,
            {},
            {"key": "value"},
            [],
            ["value"],
        ],
    )
    def test_dump_and_load(self, portable: PortableData, sut: SerializerT) -> None:
        """
        Tests :py:meth:`betty.serde.Serializer.load` and :py:meth:`betty.serde.Serializer.dump` implementations.
        """
        assert sut.load(sut.dump(portable)) == portable

    def test_load(self) -> None:
        """
        Satisfy :py:class:`betty.tests.coverage.test_coverage.TestCoverage`.
        """



from __future__ import annotations  # noqa: D100

from typing import TYPE_CHECKING, Self, final, override

from betty.app.data import AppConfiguration
from betty.argparse import assertion_to_argument_type
from betty.assertion import assert_locale
from betty.console.command import Command, CommandDefinition, CommandFunction
from betty.locale import DEFAULT_LOCALE, to_language_tag
from betty.locale.localizable.gettext import _
from betty.portable.file import assert_load_file, dump_file
from betty.service.factory import Manufacturable
from betty.service.requirement.app import require_app

if TYPE_CHECKING:
    import argparse

    from babel import Locale

    from betty.app import App


@final
@CommandDefinition("config", label=_("Configure Betty"))
class Config(Manufacturable, Command):
    """
    .. plugin:: command:config.
    """

    def __init__(self, app: App):
        self._app = app

    @override
    @classmethod
    @require_app
    async def new(cls, app: App, /) -> Self:
        return cls(app)

    @override
    async def configure(self, parser: argparse.ArgumentParser) -> CommandFunction:
        localizer = await self._app.localizer
        parser.add_argument(
            "--locale",
            default=DEFAULT_LOCALE,
            help=localizer._(
                "Set the locale for Betty's user interface. This must be an IETF BCP 47 language tag."
            ),
            type=assertion_to_argument_type(assert_locale(), localizer=localizer),
        )
        return self._command_function

    async def _command_function(self, *, locale: Locale) -> None:
        localizers = await self._app.localizers

        if AppConfiguration.FILE.exists():
            updated_configuration = AppConfiguration.data().porter.load(
                (await assert_load_file())(AppConfiguration.FILE)
            )
        else:
            updated_configuration = AppConfiguration()
        updated_configuration.locale = locale
        self._app.user.localizer = localizers.get(locale)
        await self._app.user.message_information(
            _("Betty will talk to you in {locale}").format(
                locale=locale.get_display_name() or to_language_tag(locale)
            )
        )

        await dump_file(
            AppConfiguration.data().porter.dump(updated_configuration),
            AppConfiguration.FILE,
        )



"""
Manage translations for projects.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from betty.locale.translation import (
    _new_translation,
    _update_translations,
    find_source_files,
)

if TYPE_CHECKING:
    from pathlib import Path

    from babel import Locale

    from betty.project import Project
    from betty.user import User


async def new_project_translation(
    locale: Locale, project: Project, *, user: User
) -> None:
    """
    Create a new translation for the given project.
    """
    await _new_translation(locale, project.assets_directory, user=user)


async def update_project_translations(
    project_directory_path: Path,
    source_directory_path: Path | None = None,
    exclude_source_directory_paths: set[Path] | None = None,
    *,
    _output_assets_directory_path_override: Path | None = None,
) -> None:
    """
    Update the translations for the given project.
    """
    if source_directory_path:
        source_file_paths = set(
            find_source_files(
                source_directory_path, *exclude_source_directory_paths or set()
            )
        )
    else:
        source_file_paths = set()
    await _update_translations(
        source_file_paths,
        project_directory_path / "assets",
        _output_assets_directory_path_override,
    )



"""
Documentation testing utilities.
"""

from typing import Any

import pytest

from betty.app import App
from betty.importlib import fully_qualified_name
from betty.plugin import PluginDefinition
from betty.project import Project
from betty.service.level import UNIVERSE


class PluginDocumentationTestBase:
    """
    Test a module's plugin and plugin type documentation.
    """

    _module: str

    def _match_module(self, target: Any) -> bool:
        module = getattr(target, "__module__", "")
        assert isinstance(module, str)
        return module.startswith(self._module)

    async def test(self, isolated_app: App, subtests: pytest.Subtests) -> None:
        """
        Test the plugin and plugin type documentation.
        """
        async with Project.new_isolated(isolated_app) as project, project:
            for plugin_type in UNIVERSE.plugins:
                with subtests.test():
                    self._test_plugin_type(plugin_type.type)
                async for plugin in project.plugins[plugin_type.type]:
                    with subtests.test():
                        self._test_plugin(plugin)

    def _test_plugin_type(self, plugin_type: type[PluginDefinition]) -> None:
        if not self._match_module(plugin_type):
            return
        if plugin_type.type().id.startswith("-"):
            return
        docstring = plugin_type.__doc__ or ""
        directive = f".. plugin_type:: {plugin_type.type().id}"
        assert directive in docstring, (
            f'Failed to find the "{directive}" directive in the docstring for {fully_qualified_name(plugin_type)}'
        )

    def _test_plugin(self, plugin: PluginDefinition) -> None:
        if not self._match_module(plugin.cls):
            return
        if plugin.id.startswith("-"):
            return
        docstring = plugin.cls.__doc__ or ""
        directive = f".. plugin:: {plugin.type().id}:{plugin.id}"
        assert directive in docstring, (
            f'Failed to find the "{directive}" directive in the docstring for {fully_qualified_name(plugin.cls)}'
        )



"""
Jobs.
"""

from __future__ import annotations

from random import choice
from typing import TYPE_CHECKING, override

from betty.ancestry.citation import Citation
from betty.ancestry.enclosure import Enclosure
from betty.ancestry.event import Event
from betty.ancestry.file import File
from betty.ancestry.file_reference import FileReference
from betty.ancestry.link import Link
from betty.ancestry.name import Name
from betty.ancestry.note import Note
from betty.ancestry.person import Person
from betty.ancestry.person_name import PersonName
from betty.ancestry.place import Place
from betty.ancestry.presence import Presence
from betty.ancestry.source import Source
from betty.date import Date, DateRange
from betty.dirs import ASSETS_DIRECTORY_PATH
from betty.event_type.event_types import Birth, Death, Marriage
from betty.gender.genders import Man, Woman
from betty.job import Job
from betty.locale.localizable.gettext import _
from betty.media_type.media_types import SVG
from betty.place_type.place_types import (
    Country,
    Municipality,
    Province,
    Village,
)
from betty.role.roles import Subject

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from betty.ancestry import Ancestry
    from betty.copyright_notice import CopyrightNotice
    from betty.job.scheduler import Scheduler
    from betty.license import License
    from betty.machine_name import MachineName
    from betty.service.factory import Factory


class LoadAncestry(Job):
    """
    Load the demonstration data into an ancestry.
    """

    def __init__(
        self,
        *,
        ancestry: Ancestry,
        factory: Factory,
        streetmix_copyright_notice: CopyrightNotice,
        streetmix_license: License,
    ):
        super().__init__("demo:load-ancestry")
        self._ancestry = ancestry
        self._factory = factory
        self._streetmix_copyright_notice = streetmix_copyright_notice
        self._streetmix_license = streetmix_license

    @override
    async def do(self, scheduler: Scheduler, /) -> None:
        (
            streetmix_files_per_gender,
            fallback_streetmix_files,
        ) = await self._load_streetmix_images()

        def _streetmix_image(person: Person) -> None:
            if person.file_references:
                return

            try:
                streetmix_files = streetmix_files_per_gender[person.gender.plugin().id]
            except KeyError:
                streetmix_files = fallback_streetmix_files
            streetmix_file = choice(streetmix_files)
            self._ancestry.add(FileReference(person, streetmix_file))

        netherlands = Place(
            id="betty-demo-netherlands",
            names=[
                Name(_("Netherlands")),
            ],
            links=[Link("https://en.wikipedia.org/wiki/Netherlands")],
            place_type=Country(),
        )
        self._ancestry.add(netherlands)

        north_holland = Place(
            id="betty-demo-north-holland",
            names=[
                Name(_("North Holland")),
            ],
            links=[
                Link("https://en.wikipedia.org/wiki/North_Holland"),
                Link("https://www.noord-holland.nl/"),
            ],
            place_type=Province(),
        )
        self._ancestry.add(Enclosure(enclosee=north_holland, encloser=netherlands))
        self._ancestry.add(north_holland)

        amsterdam_note = Note(
            _(
                "Did you know that while Amsterdam is the country's official capital, The Hague is the Netherlands' administrative center and seat of government?"
            )
        )

        amsterdam = Place(
            id="betty-demo-amsterdam",
            names=[
                Name(_("Amsterdam")),
            ],
            links=[
                Link("https://nl.wikipedia.org/wiki/Amsterdam"),
                Link("https://www.amsterdam.nl/"),
            ],
            notes=[amsterdam_note],
            place_type=Municipality(),
        )
        self._ancestry.add(Enclosure(enclosee=amsterdam, encloser=north_holland))
        self._ancestry.add(amsterdam)

        ilpendam = Place(
            id="betty-demo-ilpendam",
            names=[
                Name(_("Ilpendam")),
            ],
            links=[Link("https://nl.wikipedia.org/wiki/Ilpendam")],
            place_type=Village(),
        )
        self._ancestry.add(Enclosure(enclosee=ilpendam, encloser=north_holland))
        self._ancestry.add(ilpendam)

        personal_accounts = Source(
            id="betty-demo-personal-accounts",
            name=_("Personal accounts"),
        )
        self._ancestry.add(personal_accounts)

        cite_first_person_account = Citation(
            id="betty-demo-first-person-account",
            source=personal_accounts,
            location="Bart Feenstra",
        )
        self._ancestry.add(cite_first_person_account)

        noord_hollands_archief = Source(
            id="betty-demo-noord-hollands-archief",
            name="Noord-Hollands Archief",
            links=[Link("https://noord-hollandsarchief.nl/")],
        )
        self._ancestry.add(noord_hollands_archief)

        bevolkingsregister_amsterdam = Source(
            id="betty-demo-bevolkingsregister-amsterdam",
            name="Bevolkingsregister Amsterdam",
            author=_("Gemeente Amsterdam"),
            publisher=_("Gemeente Amsterdam"),
            contained_by=noord_hollands_archief,
        )
        self._ancestry.add(bevolkingsregister_amsterdam)

        david_marinus_lankester = Person(
            id="betty-demo-david-marinus-lankester", gender=Man()
        )
        _streetmix_image(david_marinus_lankester)
        self._ancestry.add(
            PersonName(
                person=david_marinus_lankester,
                individual="David Marinus",
                affiliation="Lankester",
            ),
            david_marinus_lankester,
        )

        geertruida_van_ling = Person(
            id="betty-demo-geertruida-van-ling", gender=Woman()
        )
        _streetmix_image(geertruida_van_ling)
        self._ancestry.add(
            PersonName(
                person=geertruida_van_ling,
                individual="Geertruida",
                affiliation="Van Ling",
            ),
            geertruida_van_ling,
        )

        marriage_of_dirk_jacobus_lankester_and_jannigje_palsen = Event(
            id="betty-demo-marriage-of-dirk-jacobus-lankester-and-jannigje-palsen",
            event_type=Marriage(),
            date=Date(1922, 7, 4),
            place=ilpendam,
        )
        self._ancestry.add(marriage_of_dirk_jacobus_lankester_and_jannigje_palsen)

        birth_of_dirk_jacobus_lankester = Event(
            id="betty-demo-birth-of-dirk-jacobus-lankester",
            event_type=Birth(),
            date=Date(1897, 8, 25),
            place=amsterdam,
        )
        self._ancestry.add(birth_of_dirk_jacobus_lankester)

        death_of_dirk_jacobus_lankester = Event(
            id="betty-demo-death-of-dirk-jacobus-lankester",
            event_type=Death(),
            date=Date(1986, 8, 18),
            place=amsterdam,
        )
        self._ancestry.add(death_of_dirk_jacobus_lankester)

        dirk_jacobus_lankester = Person(
            id="betty-demo-dirk-jacobus-lankester",
            gender=Man(),
            parents=[david_marinus_lankester, geertruida_van_ling],
        )
        _streetmix_image(dirk_jacobus_lankester)
        self._ancestry.add(
            PersonName(
                person=dirk_jacobus_lankester,
                individual="Dirk Jacobus",
                affiliation="Lankester",
            ),
            Presence(
                dirk_jacobus_lankester, Subject(), birth_of_dirk_jacobus_lankester
            ),
            Presence(
                dirk_jacobus_lankester, Subject(), death_of_dirk_jacobus_lankester
            ),
            Presence(
                dirk_jacobus_lankester,
                Subject(),
                marriage_of_dirk_jacobus_lankester_and_jannigje_palsen,
            ),
        )
        self._ancestry.add(dirk_jacobus_lankester)

        birth_of_marinus_david_lankester = Event(
            id="betty-demo-birth-of-marinus-david",
            event_type=Birth(),
            date=DateRange(
                Date(1874, 1, 15),
                Date(1874, 3, 21),
                start_is_boundary=True,
                end_is_boundary=True,
            ),
            place=amsterdam,
        )
        self._ancestry.add(birth_of_marinus_david_lankester)

        death_of_marinus_david_lankester = Event(
            id="betty-demo-death-of-marinus-david",
            event_type=Death(),
            date=Date(1971),
            place=amsterdam,
        )
        self._ancestry.add(death_of_marinus_david_lankester)

        marinus_david_lankester = Person(
            id="betty-demo-marinus-david-lankester",
            gender=Man(),
            parents=[david_marinus_lankester, geertruida_van_ling],
        )
        _streetmix_image(marinus_david_lankester)
        self._ancestry.add(
            PersonName(
                person=marinus_david_lankester,
                individual="Marinus David",
                affiliation="Lankester",
            ),
            Presence(
                marinus_david_lankester, Subject(), birth_of_marinus_david_lankester
            ),
            Presence(
                marinus_david_lankester, Subject(), death_of_marinus_david_lankester
            ),
        )
        self._ancestry.add(marinus_david_lankester)

        birth_of_jacoba_gesina_lankester = Event(
            id="betty-demo-birth-of-jacoba-gesina",
            event_type=Birth(),
            date=Date(1900, 3, 14),
            place=amsterdam,
        )
        self._ancestry.add(birth_of_jacoba_gesina_lankester)

        jacoba_gesina_lankester = Person(
            id="betty-demo-jacoba-gesina-lankester",
            gender=Woman(),
            parents=[david_marinus_lankester, geertruida_van_ling],
        )
        _streetmix_image(jacoba_gesina_lankester)
        self._ancestry.add(
            PersonName(
                person=jacoba_gesina_lankester,
                individual="Jacoba Gesina",
                affiliation="Lankester",
            ),
            Presence(
                jacoba_gesina_lankester, Subject(), birth_of_jacoba_gesina_lankester
            ),
        )
        self._ancestry.add(jacoba_gesina_lankester)

        jannigje_palsen = Person(id="betty-demo-jannigje-palsen", gender=Woman())
        _streetmix_image(jannigje_palsen)
        self._ancestry.add(
            PersonName(
                person=jannigje_palsen,
                individual="Jannigje",
                affiliation="Palsen",
            ),
            Presence(
                jannigje_palsen,
                Subject(),
                marriage_of_dirk_jacobus_lankester_and_jannigje_palsen,
            ),
            jannigje_palsen,
        )

        marriage_of_johan_de_boer_and_liberta_lankester = Event(
            id="betty-demo-marriage-of-johan-de-boer-and-liberta-lankester",
            event_type=Marriage(),
            date=Date(1953, 6, 19),
            place=amsterdam,
        )
        self._ancestry.add(marriage_of_johan_de_boer_and_liberta_lankester)

        cite_birth_of_liberta_lankester_from_bevolkingsregister_amsterdam = Citation(
            id="betty-demo-birth-of-liberta-lankester-from-bevolkingsregister-amsterdam",
            source=bevolkingsregister_amsterdam,
            location=_("Amsterdam"),
            date=DateRange(None, Date(2000, 1, 1), end_is_boundary=True),
        )
        self._ancestry.add(
            cite_birth_of_liberta_lankester_from_bevolkingsregister_amsterdam
        )

        birth_of_liberta_lankester = Event(
            id="betty-demo-birth-of-liberta-lankester",
            event_type=Birth(),
            date=Date(1929, 12, 22),
            place=amsterdam,
            citations=[
                cite_birth_of_liberta_lankester_from_bevolkingsregister_amsterdam
            ],
        )
        self._ancestry.add(birth_of_liberta_lankester)

        death_of_liberta_lankester = Event(
            id="betty-demo-death-of-liberta-lankester",
            event_type=Death(),
            date=Date(2015, 1, 17),
            place=amsterdam,
            citations=[cite_first_person_account],
        )
        self._ancestry.add(death_of_liberta_lankester)

        liberta_lankester_note = Note(
            _('Did you know that Liberta "Betty" Lankester is Betty\'s namesake?')
        )

        liberta_lankester = Person(
            id="betty-demo-liberta-lankester",
            gender=Woman(),
            parents=[dirk_jacobus_lankester, jannigje_palsen],
            notes=[liberta_lankester_note],
        )
        _streetmix_image(liberta_lankester)
        self._ancestry.add(
            PersonName(
                person=liberta_lankester,
                individual="Liberta",
                affiliation="Lankester",
            ),
            PersonName(
                person=liberta_lankester,
                individual="Betty",
                citations=[cite_first_person_account],
            ),
            Presence(liberta_lankester, Subject(), birth_of_liberta_lankester),
            Presence(liberta_lankester, Subject(), death_of_liberta_lankester),
            Presence(
                liberta_lankester,
                Subject(),
                marriage_of_johan_de_boer_and_liberta_lankester,
            ),
        )
        self._ancestry.add(liberta_lankester)

        birth_of_johan_de_boer = Event(
            id="betty-demo-birth-of-johan-de-boer",
            event_type=Birth(),
            date=Date(1930, 6, 20),
            place=amsterdam,
        )
        self._ancestry.add(birth_of_johan_de_boer)

        death_of_johan_de_boer = Event(
            id="betty-demo-death-of-johan-de-boer",
            event_type=Death(),
            date=Date(1999, 3, 10),
            place=amsterdam,
            citations=[cite_first_person_account],
        )
        self._ancestry.add(death_of_johan_de_boer)

        johan_de_boer = Person(id="betty-demo-johan-de-boer", gender=Man())
        _streetmix_image(johan_de_boer)
        self._ancestry.add(
            PersonName(
                person=johan_de_boer,
                individual="Johan",
                affiliation="De Boer",
            ),
            PersonName(
                person=johan_de_boer,
                individual="Hans",
            ),
            Presence(johan_de_boer, Subject(), birth_of_johan_de_boer),
            Presence(johan_de_boer, Subject(), death_of_johan_de_boer),
            Presence(
                johan_de_boer,
                Subject(),
                marriage_of_johan_de_boer_and_liberta_lankester,
            ),
            johan_de_boer,
        )

        parent_of_bart_feenstra_child_of_liberta_lankester = Person(
            id="betty-demo-parent-of-bart-feenstra-child-of-liberta-lankester",
            parents=[johan_de_boer, liberta_lankester],
        )
        _streetmix_image(parent_of_bart_feenstra_child_of_liberta_lankester)
        self._ancestry.add(
            PersonName(
                person=parent_of_bart_feenstra_child_of_liberta_lankester,
                individual="Bart's parent",
            )
        )
        self._ancestry.add(parent_of_bart_feenstra_child_of_liberta_lankester)

        birth_of_bart_feenstra = Event(
            id="betty-demo-birth-of-bart-feenstra",
            event_type=Birth(),
            date=DateRange(Date(1970, 1, 1), start_is_boundary=True),
            place=netherlands,
            citations=[cite_first_person_account],
            description=_(
                "The 'birth of the author', so to speak.",
            ),
        )
        self._ancestry.add(birth_of_johan_de_boer)

        bart_feenstra = Person(
            id="betty-demo-bart-feenstra",
            gender=Man(),
            parents=[parent_of_bart_feenstra_child_of_liberta_lankester],
        )
        Presence(bart_feenstra, Subject(), birth_of_bart_feenstra)
        _streetmix_image(bart_feenstra)
        self._ancestry.add(
            PersonName(
                person=bart_feenstra,
                individual="Bart",
                affiliation="Feenstra",
            )
        )
        self._ancestry.add(bart_feenstra)

    async def _load_streetmix_images(
        self,
    ) -> tuple[Mapping[MachineName, Sequence[File]], Sequence[File]]:
        streetmix_image_directory_path = ASSETS_DIRECTORY_PATH / "vendor" / "streetmix"
        masculine: Sequence[File] = []
        feminine: Sequence[File] = []
        androgynous: Sequence[File] = []
        file_names = [
            ("johnny-01.svg", masculine),
            ("johnny-02.svg", masculine),
            ("junebug-01.svg", feminine),
            ("junebug-02.svg", feminine),
            ("people-01.svg", feminine),
            ("people-02.svg", androgynous),
            ("people-06.svg", androgynous),
            ("people-07.svg", feminine),
            ("people-08.svg", feminine),
            ("people-09.svg", androgynous),
            ("people-11.svg", masculine),
            ("people-13.svg", feminine),
            ("people-14.svg", masculine),
            ("people-15.svg", masculine),
            ("people-16.svg", androgynous),
            ("people-17.svg", feminine),
            ("people-18.svg", feminine),
            ("people-19.svg", feminine),
            ("people-23.svg", feminine),
            ("people-24.svg", androgynous),
            ("people-31.svg", masculine),
        ]
        for file_name, appearance in file_names:
            file = File(
                streetmix_image_directory_path / file_name,
                id=f"streetmix-{file_name}",
                media_type=SVG,
                copyright_notice=self._streetmix_copyright_notice,
                license=self._streetmix_license,
            )
            appearance.append(file)
            self._ancestry.add(file)

        return {
            Woman.plugin().id: feminine + androgynous,
            Man.plugin().id: masculine + androgynous,
        }, androgynous



"""
Service levels.
"""

from __future__ import annotations

from collections import defaultdict
from importlib import metadata
from typing import TYPE_CHECKING, Any, Final

from betty.collection.keyed.adapter import KeyedCollectionAdapter
from betty.collection.keyed.error import ErroringKeyedCollection
from betty.machine_name import MachineName
from betty.plugin import PluginDefinition
from betty.plugin.error import PluginTypeNotFound
from betty.service.provider import service

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping

    from betty.collection.keyed import KeyedCollection
    from betty.plugin.discovery import ResolvableDiscovery
    from betty.service.factory import Factory
    from betty.service.plugin import PluginManager


class _PluginTypeNotFound(PluginTypeNotFound, KeyError):
    pass


class ServiceLevel:
    """
    A service level.
    """

    def __init__(
        self,
        *args: Any,
        plugins: Mapping[
            type[PluginDefinition], Iterable[ResolvableDiscovery[PluginDefinition]]
        ]
        | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self._plugin_discovery = defaultdict(
            lambda: None, {} if plugins is None else plugins
        )

    @service
    def factory(self) -> Factory:
        """
        The object factory.
        """
        from betty.service.factory import Factory

        return Factory(self)

    @service
    def plugins(
        self,
    ) -> KeyedCollection[
        type[PluginDefinition],
        type[PluginDefinition] | MachineName | str,
        PluginManager,
    ]:
        """
        The available plugin types and plugins.
        """
        from betty.service.plugin import PluginManager

        def _resolve_plugin_type_id(
            key: type[PluginDefinition] | MachineName | str,
        ) -> MachineName:
            return (
                key.id
                if isinstance(key, PluginDefinition)
                else key.type().id
                if isinstance(key, type) and issubclass(key, PluginDefinition)
                else MachineName.resolve(key)
            )

        plugin_types = [
            entry_point.load()
            for entry_point in metadata.entry_points(group="betty.plugin")
        ]
        plugin_types.extend(self._plugin_discovery.keys())
        return ErroringKeyedCollection(
            KeyedCollectionAdapter(
                {
                    plugin_type.type().id: PluginManager(
                        self, plugin_type, self._plugin_discovery[plugin_type]
                    )
                    for plugin_type in plugin_types
                },  # ty:ignore[invalid-argument-type]
                key_resolver=_resolve_plugin_type_id,
            ),
            lambda error, key: _PluginTypeNotFound(
                _resolve_plugin_type_id(key), [x.type.type().id for x in self.plugins]
            ),
        )


UNIVERSE: Final[ServiceLevel] = ServiceLevel()
"""
The universal service level.
"""



"""
The HTTP client API.
"""

from aiohttp import ClientError
from aiohttp.client_middlewares import ClientHandlerType
from aiohttp.client_reqrep import ClientRequest, ClientResponse

from betty.user import User


class ClientErrorToUserMessageMiddleware:
    """
    Log client errors to a user.
    """

    def __init__(self, user: User):
        self._user = user

    async def __call__(
        self, request: ClientRequest, handler: ClientHandlerType
    ) -> ClientResponse:
        """
        Call the middleware.
        """
        try:
            return await handler(request)
        except ClientError as error:
            await self._user.message_debug(str(error))
            raise



"""Define Betty's core application functionality."""

from __future__ import annotations

from contextlib import AsyncExitStack, asynccontextmanager
from os import environ
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, Self, final, override

from aiofiles.tempfile import TemporaryDirectory
from aiohttp_client_cache.backends.filesystem import FileBackend
from aiohttp_client_cache.session import CachedSession

import betty
import betty.dirs
from betty.app.data import AppConfiguration
from betty.asset import AssetRepository, StaticAssetRepository
from betty.cache.file import BinaryFileCache, PickledFileCache
from betty.cache.no_op import NoOpCache
from betty.dirs import CACHE_DIRECTORY_PATH
from betty.http_client import ClientErrorToUserMessageMiddleware
from betty.http_client.rate_limit import RateLimitDefinition, RateLimitMiddleware
from betty.life_cycle import LifeCycle
from betty.life_cycle.manage import ManagedLifeCycle
from betty.locale import DEFAULT_LOCALE, ResolvableLocale, resolve_locale
from betty.locale.localize import Localizer, LocalizerRepository
from betty.locale.translation import (
    DEFAULT_TRANSLATION_REPOSITORY,
    AssetTranslationRepository,
    TranslationRepository,
)
from betty.multiprocessing import ProcessPoolExecutor
from betty.plugin.ordered import sort_ordered_plugin_graph
from betty.portable.file import assert_load_file
from betty.service.factory import DataManufacturable
from betty.service.level import UNIVERSE, ServiceLevel
from betty.service.provider import ServiceFactory, service
from betty.typing import threadsafe
from betty.user.no_op import NoOpUser

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Iterable, Mapping
    from concurrent import futures

    import aiohttp

    from betty.cache import Cache
    from betty.plugin import PluginDefinition
    from betty.plugin.discovery import ResolvableDiscovery
    from betty.user import User


@final
@threadsafe
class App(DataManufacturable[AppConfiguration], ServiceLevel, ManagedLifeCycle):
    """
    The Betty application.

    .. list-table::
       :widths: 20 10
       :header-rows: 0

       * - Configuration
         - :py:class:`betty.app.data.AppConfiguration`
    """

    def __init__(
        self,
        *,
        cache_directory: Path | None = None,
        cache_factory: ServiceFactory[App, Cache[Any]] | None = None,
        locale: ResolvableLocale | None = None,
        plugins: Mapping[
            type[PluginDefinition], Iterable[ResolvableDiscovery[PluginDefinition]]
        ]
        | None = None,
        process_pool: futures.ProcessPoolExecutor | None = None,
        translations: TranslationRepository | None = None,
        user: User | None = None,
    ):
        from betty.rich.user import RichUser

        cls = type(self)
        super().__init__(plugins=plugins)
        self.life_cycle.on_bootstrap(self._bootstrap_localizer)
        self._locale = DEFAULT_LOCALE if locale is None else resolve_locale(locale)
        self._user = user or RichUser()
        if isinstance(self._user, LifeCycle):
            self.life_cycle.attach(self._user)
        if process_pool is not None:
            cls.process_pool.override(self, process_pool)
        if translations is not None:
            cls.translations.override(self, translations)
        self._cache_directory = (
            Path(environ.get("BETTY_CACHE_DIRECTORY", CACHE_DIRECTORY_PATH))
            if cache_directory is None
            else cache_directory
        )
        cls.cache.override_factory(
            self,
            (lambda _: PickledFileCache[Any](self._cache_directory))
            if cache_factory is None
            else cache_factory,
        )

    async def _bootstrap_localizer(self) -> None:
        self._user.localizer = await self.localizer

    @override
    @classmethod
    def new_data_cls(cls) -> type[AppConfiguration]:
        return AppConfiguration

    @override
    @classmethod
    async def new(cls, services: ServiceLevel, data: AppConfiguration, /) -> Self:
        return cls(locale=data.locale)

    @classmethod
    async def new_from_environment(cls) -> Self:
        """
        Create a new application from the environment.
        """
        if AppConfiguration.FILE.exists():
            configuration = AppConfiguration.data().porter.load(
                (await assert_load_file())(AppConfiguration.FILE)
            )
            return await cls.new(UNIVERSE, configuration)
        return cls()

    @classmethod
    @asynccontextmanager
    async def new_isolated(
        cls,
        *,
        cache_directory: Path | None = None,
        cache_factory: ServiceFactory[App, Cache[Any]] | None = None,
        plugins: Mapping[
            type[PluginDefinition], Iterable[ResolvableDiscovery[PluginDefinition]]
        ]
        | None = None,
        process_pool: futures.ProcessPoolExecutor | None = None,
        user: User | None = None,
        translations: TranslationRepository | None | Literal[False] = False,
    ) -> AsyncIterator[Self]:
        """
        Create a new, isolated, temporary application.

        The application will not use any persistent caches, or leave
        any traces on the system.
        """
        async with AsyncExitStack() as exit_stack:
            if cache_directory is None:
                cache_directory = Path(
                    await exit_stack.enter_async_context(TemporaryDirectory())
                )
            yield cls(
                cache_directory=cache_directory,
                cache_factory=(lambda _: NoOpCache())
                if cache_factory is None
                else cache_factory,
                plugins=plugins,
                process_pool=process_pool,
                user=NoOpUser() if user is None else user,
                translations=DEFAULT_TRANSLATION_REPOSITORY
                if translations is False
                else translations,
            )

    @property
    def user(self) -> User:
        """
        The current user session.
        """
        return self._user

    @service
    def assets(self) -> AssetRepository:
        """
        The assets file system.
        """
        return StaticAssetRepository(betty.dirs.ASSETS_DIRECTORY_PATH)

    @service
    async def translations(self) -> TranslationRepository:
        """
        The available translations.
        """
        translations = AssetTranslationRepository(self.assets, self.binary_file_cache)
        await translations.bootstrap()
        return translations

    @service
    async def localizer(self) -> Localizer:
        """
        Get the application's user-facing localizer.
        """
        return (await self.localizers).get(self._locale)

    @service
    async def localizers(self) -> LocalizerRepository:
        """
        The available localizers.
        """
        return LocalizerRepository(await self.translations)

    @service
    async def http_client(self) -> aiohttp.ClientSession:
        """
        The HTTP client.
        """
        http_rate_limits = self.plugins[RateLimitDefinition]
        rate_limit_sorter = await sort_ordered_plugin_graph(
            http_rate_limits, [x async for x in http_rate_limits]
        )

        http_client: aiohttp.ClientSession = CachedSession(
            cache=FileBackend(self.binary_file_cache.with_scope("http-client").path),
            headers={
                "User-Agent": "Betty (https://betty.readthedocs.io/)",
            },
            middlewares=[
                ClientErrorToUserMessageMiddleware(self.user),
                RateLimitMiddleware(
                    [
                        await self.factory.new(
                            (await http_rate_limits[rate_limit_id]).cls
                        )
                        for rate_limit_id in rate_limit_sorter.static_order()
                    ]
                ),
            ],
        )

        self.life_cycle.on_shutdown(lambda wait: http_client.close())

        return http_client

    @service
    def cache(self) -> Cache[Any]:
        """
        The cache.
        """
        raise Exception(
            "This must never happen, because a cache must be set explicitly."
        )

    @service
    def binary_file_cache(self) -> BinaryFileCache:
        """
        The binary file cache.
        """
        return BinaryFileCache(self._cache_directory)

    @service
    def process_pool(self) -> futures.ProcessPoolExecutor:
        """
        The shared process pool.

        Use this to run CPU/computationally-heavy tasks in other processes.
        """
        process_pool = ProcessPoolExecutor()
        self.life_cycle.on_shutdown(
            lambda wait: process_pool.shutdown(wait, cancel_futures=not wait)
        )
        return process_pool



"""
Manage translations for project extensions.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from betty.exception import HumanFacingException
from betty.locale.localizable.gettext import _
from betty.locale.translation import (
    _new_translation,
    _update_translations,
    find_source_files,
)

if TYPE_CHECKING:
    from pathlib import Path

    from babel import Locale

    from betty.extension import ExtensionDefinition
    from betty.user import User


def assert_extension_assets_directory_path(extension: ExtensionDefinition) -> Path:
    """
    Check that the given extension has an assets directory, and return its path.
    """
    assets_directory_path = extension.assets_directory
    if assets_directory_path is None:
        raise HumanFacingException(
            _("{extension} does not have an assets directory.").format(
                extension=extension.id
            )
        )
    return assets_directory_path


async def new_extension_translation(
    locale: Locale, extension: ExtensionDefinition, *, user: User
) -> None:
    """
    Create a new translation for the given extension.
    """
    await _new_translation(
        locale, assert_extension_assets_directory_path(extension), user=user
    )


async def update_extension_translations(
    extension: ExtensionDefinition,
    source_directory_path: Path | None = None,
    exclude_source_directory_paths: set[Path] | None = None,
    *,
    _output_assets_directory_path_override: Path | None = None,
) -> None:
    """
    Update the translations for the given extension.
    """
    if source_directory_path:
        source_file_paths = set(
            find_source_files(
                source_directory_path, *exclude_source_directory_paths or set()
            )
        )
    else:
        source_file_paths = set()
    await _update_translations(
        source_file_paths,
        assert_extension_assets_directory_path(extension),
        _output_assets_directory_path_override,
    )



"""
Test utilities for :py:mod:`betty.ancestry.has_links`.
"""

from betty.ancestry.has_links import HasLinks
from betty.model import EntityDefinition
from betty.test_utils.locale.localizable import (
    DUMMY_COUNTABLE_LOCALIZABLE,
    DUMMY_LOCALIZABLE,
)


@EntityDefinition(
    "dummy-has-links",
    label=DUMMY_LOCALIZABLE,
    label_plural=DUMMY_LOCALIZABLE,
    label_countable=DUMMY_COUNTABLE_LOCALIZABLE,
)
class DummyHasLinks(HasLinks):
    """
    A dummy :py:class:`betty.ancestry.has_links.HasLinks` entity.
    """



"""
Class-based definitions.
"""

from typing import Any

from betty.importlib import fully_qualified_name


class ClsDefinition[ClsT = Any]:
    """
    A definition for a Python class.
    """

    def __init__(self, *args: Any, cls: type[ClsT] | None = None, **kwargs: Any):
        self._cls: type[ClsT] | None = None
        if cls is not None:
            self._set_cls(cls)

    @property
    def cls(self) -> type[ClsT]:
        """
        The class.

        :raises ValueError: Raised if the definition was not yet used to decorate a class.
        """
        if self._cls is None:
            raise ValueError("This definition does not yet have a class.")
        assert self._cls is not None
        return self._cls

    def __call__(self, cls: type[ClsT]) -> type[ClsT]:
        """
        Decorate a class and set it on this definition.

        :raises ValueError: Raised if the definition was already used to decorate a class.
        """
        self._set_cls(cls)
        return cls

    def _set_cls(self, cls: type[ClsT]) -> None:
        if self._cls is not None:
            raise ValueError(
                f"This definition already has a class: {fully_qualified_name(self._cls)}."
            )
        assert self._cls is None
        self._cls = cls



"""
Provide the demonstration project.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from betty.ancestry.event import Event
from betty.ancestry.person import Person
from betty.ancestry.place import Place
from betty.ancestry.source import Source
from betty.content import ContentManufacturer
from betty.content.contents import Render, RenderConfiguration
from betty.extension import ExtensionManufacturer
from betty.extension.demo.content import _IncompleteTranslationWarning
from betty.extension.raspberry_mint import Breakpoint, RaspberryMint
from betty.extension.raspberry_mint.content import (
    Columns,
    ColumnsConfiguration,
    EntityCard,
    Section,
    SectionConfiguration,
)
from betty.extension.raspberry_mint.data import RaspberryMintConfiguration
from betty.extension.raspberry_mint.default import regional_content
from betty.license import LicenseManufacturer
from betty.locale import DEFAULT_LOCALE
from betty.locale.localizable.gettext import _
from betty.locale.localizable.markup import Chain
from betty.locale.localize import LocalizerRepository
from betty.media_type.media_types import HTML
from betty.model.reference import EntityReference
from betty.project import Project
from betty.project.data import ProjectConfiguration

if TYPE_CHECKING:
    from pathlib import Path

    from betty.app import App


async def create_project(app: App, project_directory_path: Path) -> Project:
    """
    Create a new demonstration project.
    """
    from betty.extension.demo import Demo

    translations = await app.translations
    localizer_repository = LocalizerRepository(translations)
    localizers = [localizer_repository.get(locale) for locale in translations.locales]

    configuration = ProjectConfiguration(
        name=Demo.plugin().id,
        license=LicenseManufacturer("spdx-gpl-3--0-or-later"),
        title=_("A Betty demonstration"),
        author=_("Bart Feenstra and contributors"),
        url="https://example.com",
        extensions=[
            ExtensionManufacturer(Demo),
            ExtensionManufacturer(
                RaspberryMint,
                RaspberryMintConfiguration(
                    regional_content={
                        **regional_content(localizers=localizers),
                        "front-page-content": [
                            ContentManufacturer(
                                Columns,
                                ColumnsConfiguration([[_IncompleteTranslationWarning]]),
                            ),
                            ContentManufacturer(
                                Section,
                                SectionConfiguration(
                                    ContentManufacturer(
                                        Columns,
                                        ColumnsConfiguration(
                                            [
                                                [
                                                    ContentManufacturer(
                                                        Render,
                                                        RenderConfiguration(
                                                            Chain(
                                                                "<h2>",
                                                                _("Get started"),
                                                                "</h2>"
                                                                '<a href="https://betty.readthedocs.io/" class="view-more">',
                                                                _(
                                                                    "Read the documentation"
                                                                ),
                                                                "</a>",
                                                                '<a href="https://github.com/bartfeenstra/betty/" class="view-more">',
                                                                _("View the code"),
                                                                "</a>",
                                                            ),
                                                            HTML,
                                                        ),
                                                    ),
                                                ],
                                                [
                                                    ContentManufacturer(
                                                        Render,
                                                        RenderConfiguration(
                                                            Chain(
                                                                "<p>",
                                                                _(
                                                                    "Betty was named after <a href=\"betty-entity://person/betty-demo-liberta-lankester\">Liberta 'Betty' Lankester</a>, and this website shows a small sample of her family history. You can browse the pages about her and some of her family to get an idea of what a Betty site looks like."
                                                                ),
                                                                "</p>",
                                                            ),
                                                            HTML,
                                                        ),
                                                    ),
                                                ],
                                            ],
                                            width={
                                                Breakpoint.XS: [12, 12],
                                                Breakpoint.MD: [5, 6],
                                                Breakpoint.LG: [4, 7],
                                            },
                                        ),
                                    ),
                                    heading=_("Welcome"),
                                    visually_hide_heading=True,
                                ),
                            ),
                            ContentManufacturer(
                                Section,
                                SectionConfiguration(
                                    ContentManufacturer(
                                        Columns,
                                        ColumnsConfiguration(
                                            [
                                                [
                                                    ContentManufacturer(
                                                        EntityCard,
                                                        EntityReference(
                                                            Place,
                                                            "betty-demo-amsterdam",
                                                        ),
                                                    )
                                                ],
                                                [
                                                    ContentManufacturer(
                                                        EntityCard,
                                                        EntityReference(
                                                            Person,
                                                            "betty-demo-liberta-lankester",
                                                        ),
                                                    )
                                                ],
                                                [
                                                    ContentManufacturer(
                                                        EntityCard,
                                                        EntityReference(
                                                            Place,
                                                            "betty-demo-netherlands",
                                                        ),
                                                    )
                                                ],
                                            ],
                                            width={
                                                Breakpoint.XS: [12, 12, 12],
                                                Breakpoint.MD: [6, 6, 6],
                                                Breakpoint.LG: [4, 4, 4],
                                            },
                                        ),
                                    ),
                                    heading=_("Explore a family history..."),
                                ),
                            ),
                        ],
                        "front-page-summary": [
                            ContentManufacturer(
                                Render,
                                RenderConfiguration(
                                    _(
                                        "Betty is an application that takes a family tree and builds a website out of it, much like the one you are viewing right now. The more information your genealogical research contains, the more interactivity Betty can add to your site, such as media galleries, maps, and browsable family trees."
                                    )
                                ),
                            ),
                        ],
                    }
                ),
            ),
        ],
        entity_types=[
            Person,
            Event,
            Place,
            Source,
        ],
        locales=[
            # The first configured locale is the project default.
            DEFAULT_LOCALE,
            *[locale for locale in translations.locales if locale != DEFAULT_LOCALE],
        ],
    )
    return Project(
        app, project_directory_path / "betty.json", configuration=configuration
    )



"""
Dynamic content.
"""

from typing import Self, override

from betty.ancestry.has_links import HasLinks
from betty.content import ContentDefinition
from betty.content.contents import Template, TemplateBuild
from betty.document import Document
from betty.extension.wiki import Wiki
from betty.locale.localizable.gettext import _
from betty.project import Project
from betty.service.factory import Manufacturable
from betty.service.requirement.extension import require_extension
from betty.service.requirement.project import require_project


@ContentDefinition("wiki-wikipedia-summary", label=_("Wikipedia summary"))
class WikipediaSummary(Template, Manufacturable):
    """
    A Wikipedia summary.

    .. plugin:: content:wiki-wikipedia-summary
    """

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        await require_extension(Wiki, project)
        return cls(jinja=await project.jinja)

    @override
    async def build_template(self, document: Document) -> TemplateBuild:
        if isinstance(document.resource, HasLinks):
            return "component/wiki/wikipedia-summary.html.j2", {
                "links": document.resource.links
            }
        return None



"""
Test utilities for :py:mod:`betty.content`.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final, override

from betty.content import Content, ContentDefinition

if TYPE_CHECKING:
    from betty.document import Document


@final
@ContentDefinition("no-op", label="No-op")
class NoOpContent(Content):
    """
    A content plugin that never provides any content.
    """

    @override
    async def build(self, *, document: Document) -> str | None:
        return None



"""
Logging utilities.
"""

import contextlib
import logging
import threading
from asyncio import (
    CancelledError,
    get_running_loop,
    run_coroutine_threadsafe,
    to_thread,
)
from collections.abc import Callable, Coroutine
from functools import partial
from queue import Empty, Queue
from time import sleep
from typing import final, override

from betty.functools import Result, ResultUnavailable, suppress
from betty.life_cycle.manage import ManagedLifeCycle
from betty.user import User


@final
class UserHandler(ManagedLifeCycle, logging.Handler):
    """
    Output log records through a :py:class`betty.user.User`.
    """

    def __init__(self, user: User, /):
        super().__init__()
        self._user = user
        self._result = Result(self._consume)
        self._thread = threading.Thread(
            name=self.__class__.__name__, target=suppress(self._result, BaseException)
        )
        self.life_cycle.on((self._thread.start, self._shutdown_thread))
        self._queue = Queue[Callable[[], Coroutine[None, None, None]]]()
        self._finish = threading.Event()
        self._loop = get_running_loop()

    async def _shutdown_thread(self, *, wait: bool = True) -> None:
        self._finish.set()
        with contextlib.suppress(CancelledError):
            await to_thread(self._thread.join)
        # If no log messages were recorded, there is no result.
        with contextlib.suppress(ResultUnavailable):
            self._result.result()

    def _consume(self) -> None:
        final_iteration = False
        while True:
            try:
                task = self._queue.get_nowait()
            except Empty:
                if self._finish.is_set():
                    # Perform one final iteration to account for race conditions between the finish event being set and
                    # the final tasks being added to the queue.
                    if final_iteration:
                        return
                    final_iteration = True
                # Sleep to prevent the loop from taking up CPU time when the queue is empty.
                sleep(0.001)
            else:
                run_coroutine_threadsafe(task(), self._loop).result()

    @override
    def emit(self, record: logging.LogRecord) -> None:
        self._queue.put_nowait(partial(self._user.message_log, record))



"""
Provide Betty's default Jinja2 filters.
"""

from __future__ import annotations

import json as stdjson
import re
import warnings
from asyncio import gather, get_running_loop, run
from contextlib import suppress
from io import BytesIO
from typing import TYPE_CHECKING, Any
from urllib.parse import quote

import aiofiles
from aiofiles.os import makedirs
from babel.dates import format_date
from geopy import units
from geopy.format import DEGREES_FORMAT
from jinja2 import pass_context
from jinja2.async_utils import auto_aiter
from markupsafe import Markup
from pdf2image.pdf2image import convert_from_path
from PIL import Image
from PIL.Image import DecompressionBombWarning

from betty.ancestry.file import File
from betty.ancestry.file_reference import FileReference
from betty.content import Content, ContentDefinition, ContentManufacturer, build
from betty.hashid import hashid_file_meta
from betty.image import FocusArea, Size, image_file_path_format, resize_cover
from betty.locale import HasLocaleStr, ResolvableLocale, to_language_tag
from betty.media_type import MediaType
from betty.media_type.media_types import HTML, SVG
from betty.os import link_or_copy

if TYPE_CHECKING:
    import datetime
    from collections.abc import (
        AsyncIterator,
        Awaitable,
        Callable,
        Iterable,
        Iterator,
        Mapping,
    )
    from pathlib import Path

    from jinja2.runtime import Context

    from betty.ancestry.date import HasDate
    from betty.date import ResolvableDate
    from betty.locale.localizable import Localizable
    from betty.plugin.factory import ResolvablePluginManufacturer


@pass_context
async def filter_url(
    context: Context,
    resource: Any,
    locale: ResolvableLocale | None = None,
    media_type: str | None = None,
    **kwargs: Any,
) -> str:
    """
    Generate a URL for a resource.
    """
    from betty.jinja import context_localizer, context_project

    url_generator = await context_project(context).url_generator
    return url_generator.generate(
        resource,
        media_type=MediaType(media_type) if media_type else HTML,
        locale=locale or context_localizer(context).locale,
        **kwargs,
    )


@pass_context
def filter_localize(
    context: Context,
    localizable: Localizable,
) -> str:
    """
    Localize a value using the context's current localizer.
    """
    from betty.jinja import context_localizer

    return localizable.localize(context_localizer(context))


_CHARACTER_ORDERTO_HTML_LANG_MAP = {
    "left-to-right": "ltr",
    "right-to-left": "rtl",
}


@pass_context
def filter_html_lang(context: Context, has_locale: str) -> str | Markup:
    """
    Optionally add the necessary HTML to indicate the localized string has a different locale than the surrounding HTML.
    """
    from betty.jinja import context_localizer

    if not isinstance(has_locale, HasLocaleStr):
        return has_locale

    localizer = context_localizer(context)
    result: str | Markup = has_locale
    if has_locale.locale != localizer.locale:
        localizer_dir = _CHARACTER_ORDERTO_HTML_LANG_MAP[
            localizer.locale.character_order
        ]
        if has_locale.locale is None:
            has_locale_dir = "auto"
        else:
            has_locale_dir = _CHARACTER_ORDERTO_HTML_LANG_MAP[
                has_locale.locale.character_order
            ]
        dir_attribute = (
            f' dir="{has_locale_dir}"' if has_locale_dir != localizer_dir else ""
        )
        result = f'<span lang="{to_language_tag(has_locale.locale)}"{dir_attribute}>{has_locale}</span>'
    if context.eval_ctx.autoescape:
        result = Markup(result)
    return result


def filter_json_dump(data: Any, indent: int | None = None) -> str:
    """
    Dump a value to a JSON string.
    """
    return stdjson.dumps(data, indent=indent)


def filter_json_load(data: str) -> Any:
    """
    Load a value from a JSON string.
    """
    return stdjson.loads(data)


_paragraph_re = re.compile(r"(?:\r\n|\r|\n){2,}")


def filter_format_degrees(degrees: int) -> str:
    """
    Format geographic coordinates.
    """
    arcminutes = units.arcminutes(degrees=degrees - int(degrees))
    arcseconds = units.arcseconds(arcminutes=arcminutes - int(arcminutes))
    format_dict = {
        "deg": "°",
        "arcmin": "'",
        "arcsec": '"',
        "degrees": degrees,
        "minutes": round(abs(arcminutes)),
        "seconds": round(abs(arcseconds)),
    }
    return DEGREES_FORMAT % format_dict


async def filter_unique[T](values: Iterable[T]) -> AsyncIterator[T]:
    """
    Iterate over an iterable of values and only yield those values that have not been yielded before.
    """
    seen = []
    async for value in auto_aiter(values):
        if value not in seen:
            yield value
            seen.append(value)


@pass_context
async def filter_file(context: Context, file: File) -> str:
    """
    Preprocess a file for use in a page.

    :return: A ``betty-static://`` URL resource from which a public URL can be generated.
    """
    from betty.jinja import context_context, context_project

    project = context_project(context)
    job_context = context_context(context)

    execute_filter = True
    if job_context:
        job_cache_item_id = f"filter_file:{file.id}"
        async with job_context.cache.hasset(job_cache_item_id) as setter:
            if setter:
                await setter(True)
            else:
                execute_filter = False
    if execute_filter:
        file_destination_path = (
            project.www_directory / "file" / file.id / "file" / file.name
        )
        await makedirs(file_destination_path.parent, exist_ok=True)
        await link_or_copy(file.path, file_destination_path)

    return f"betty-static:///file/{quote(file.id)}/file/{quote(file.name)}"


@pass_context
async def filter_image_resize_cover(
    context: Context,
    filey: File | FileReference,
    size: Size | None = None,
    *,
    focus: FocusArea | None = None,
) -> str:
    """
    Preprocess an image file for use in a page.

    :return: A ``betty-static://`` URL resource from which a public URL can be generated.
    """
    from betty.jinja import context_context, context_project

    file = filey if isinstance(filey, File) else filey.file
    assert file is not None
    file_reference = filey if isinstance(filey, FileReference) else None

    if (
        focus is None
        and file_reference is not None
        and file_reference.focus is not None
    ):
        focus = file_reference.focus

    # Treat SVGs as regular files.
    if file.media_type and file.media_type == SVG:
        return await filter_file(context, file)

    project = context_project(context)
    job_context = context_context(context)

    destination_name = f"{file.id}-"
    if size is not None:
        width, height = size
        if width is None:
            destination_name += f"-x{height}"
        elif height is None:
            destination_name += f"{width}x-"
        else:
            destination_name += f"{width}x{height}"
    if focus is not None:
        destination_name += f"-{focus[0]}x{focus[1]}x{focus[2]}x{focus[3]}"

    file_directory_path = project.www_directory / "file"

    if file.media_type:
        if file.media_type.type == "image":
            image_loader = _load_image_image
            destination_name += file.path.suffix
        elif file.media_type.type == "application" and file.media_type.subtype == "pdf":
            image_loader = _load_image_application_pdf
            destination_name += "." + "jpg"
        else:
            raise ValueError(
                f'Cannot convert a file of media type "{file.media_type}" to an image.'
            )
    else:
        raise ValueError("Cannot convert a file without a media type to an image.")

    cache_item_id = f"{await hashid_file_meta(file.path)}:{destination_name}"
    execute_filter = True
    if job_context:
        async with job_context.cache.with_scope("filter_image").hasset(
            cache_item_id
        ) as setter:
            if setter:
                await setter(True)
            else:
                execute_filter = False
    if execute_filter:
        loop = get_running_loop()
        await loop.run_in_executor(
            project.app.process_pool,
            _execute_filter_image,
            image_loader,
            file.path,
            project.app.binary_file_cache.with_scope("image").cache_item_file_path(
                cache_item_id
            ),
            file_directory_path,
            destination_name,
            size,
            focus,
        )
    return f"betty-static:///file/{quote(destination_name)}"


async def _load_image_image(file_path: Path) -> Image.Image:
    # We want to read the image asynchronously and prevent Pillow from keeping too many file
    # descriptors open simultaneously, so we read the image ourselves and store the contents
    # in a synchronous file object.
    async with aiofiles.open(file_path, "rb") as f:
        image_f = BytesIO(await f.read())
    # Ignore warnings about decompression bombs, because we know where the files come from.
    with warnings.catch_warnings(action="ignore", category=DecompressionBombWarning):
        return Image.open(image_f, formats=[image_file_path_format(file_path)])


async def _load_image_application_pdf(file_path: Path) -> Image.Image:
    # Ignore warnings about decompression bombs, because we know where the files come from.
    with warnings.catch_warnings(action="ignore", category=DecompressionBombWarning):
        return convert_from_path(file_path)[0]


def _execute_filter_image(
    image_loader: Callable[[Path], Awaitable[Image.Image]],
    file_path: Path,
    cache_item_file_path: Path,
    destination_directory_path: Path,
    destination_name: str,
    size: Size | None,
    focus: FocusArea | None,
) -> None:
    run(
        __execute_filter_image(
            image_loader,
            file_path,
            cache_item_file_path,
            destination_directory_path,
            destination_name,
            size,
            focus,
        )
    )


async def __execute_filter_image(
    image_loader: Callable[[Path], Awaitable[Image.Image]],
    file_path: Path,
    cache_item_file_path: Path,
    destination_directory_path: Path,
    destination_name: str,
    size: Size | None,
    focus: FocusArea | None,
) -> None:
    destination_file_path = destination_directory_path / destination_name
    await makedirs(destination_directory_path, exist_ok=True)

    # If no customizations are needed, work straight from the source.
    if size is None and file_path.suffix == destination_file_path.suffix:
        await link_or_copy(file_path, destination_file_path)
        return

    try:
        # Try using a previously cached image.
        await link_or_copy(cache_item_file_path, destination_file_path)
    except FileNotFoundError:
        # Apply customizations, and cache the customized image.
        original_image = converted_image = await image_loader(file_path)
        try:
            await makedirs(cache_item_file_path.parent, exist_ok=True)
            if size is not None:
                converted_image = resize_cover(converted_image, size, focus=focus)
            converted_image.save(
                cache_item_file_path,
                format=image_file_path_format(destination_file_path),
            )
            del converted_image
        finally:
            original_image.close()
            del original_image
        await link_or_copy(cache_item_file_path, destination_file_path)


@pass_context
def filter_negotiate_has_dates(
    context: Context, has_dates: Iterable[HasDate], date: ResolvableDate | None
) -> HasDate | None:
    """
    Try to find an object whose date falls in the given date.

    :param date: A date to select by. If ``None``, then today's date is used.
    """
    with suppress(StopIteration):
        return next(filter_select_has_dates(context, has_dates, date))
    return None


@pass_context
def filter_select_has_dates(
    context: Context, has_dates: Iterable[HasDate], date: ResolvableDate | None
) -> Iterator[HasDate]:
    """
    Select all objects whose date falls in the given date.

    :param date: A date to select by. If ``None``, then today's date is used.
    """
    if date is None:
        date = context.resolve_or_missing("today")
    return filter(
        lambda dated: (
            dated.date is None or dated.date.comparable and dated.date in date
        ),
        has_dates,
    )


@pass_context
async def filter_build_content(
    context: Context,
    contents: Iterable[ResolvablePluginManufacturer[ContentDefinition, Content]],
) -> Markup:
    """
    Build content from content configuration.
    """
    from betty.jinja import context_document, context_project

    project = context_project(context)
    return await build(
        context_document(context),
        await gather(
            *map(
                project.factory.new,
                ContentManufacturer.resolve_sequence(contents),
            )
        ),
    ) or Markup("")


@pass_context
def filter_format_datetime_datetime(
    context: Context, datetime_datetime: datetime.datetime, /
) -> str:
    """
    Format a datetime date to a human-readable string.
    """
    from betty.jinja import context_localizer

    localizer = context_localizer(context)
    return format_date(datetime_datetime, "long", locale=localizer.locale)


async def filters() -> Mapping[str, Callable[..., Any]]:
    """
    Define the available filters.
    """
    return {
        "file": filter_file,
        "format_datetime_datetime": filter_format_datetime_datetime,
        "format_degrees": filter_format_degrees,
        "image_resize_cover": filter_image_resize_cover,
        "html_lang": filter_html_lang,
        "json_dump": filter_json_dump,
        "json_load": filter_json_load,
        "localize": filter_localize,
        "negotiate_has_dates": filter_negotiate_has_dates,
        "build_content": filter_build_content,
        "select_has_dates": filter_select_has_dates,
        "to_language_tag": to_language_tag,
        "unique": filter_unique,
        "url": filter_url,
    }



"""
Test utilities for :py:mod:`betty.data`.
"""

from typing import Any

import pytest

from betty.data import Data
from betty.data.aggregate.record.object import ObjectDefinition
from betty.data.str import StrDefinition
from betty.importlib import fully_qualified_name
from betty.locale.localizable.plain import Plain
from betty.locale.localize import DEFAULT_LOCALIZER
from betty.property import Optional, Property


class DataTestBase[DataT: Data]:
    """
    A base class for testing :py:class:`betty.data.Data` subclasses.
    """

    sut_cls: type[DataT]
    """
    The system under test.
    """

    def test_data(self) -> None:
        """
        Tests :py:meth:`betty.data.Data.data` implementations.
        """
        self.sut_cls.data()

    def test_cls_docstring(self) -> None:
        """
        Test the class's docstring.
        """
        docstring = self.sut_cls.__doc__
        assert docstring, "Failed asserting that the class has a docstring"
        directive = f".. data:: {fully_qualified_name(self.sut_cls)}"
        assert directive in docstring, (
            f"Failed to find `{directive}` in the class's docstring"
        )

    def test_data__samples__should_provide_at_least_one(self) -> None:
        """
        Tests that the data definition provides at least one sample.
        """
        assert len(list(self.sut_cls.data().samples)), (
            "Failed asserting that at least one sample is provided"
        )

    def test_data__porter__with_samples(self, subtests: pytest.Subtests) -> None:
        """
        Tests that the data definition can consistently dump and load its samples.
        """
        samples = list(self.sut_cls.data().samples)
        for sample in samples:
            with subtests.test(str(sample.label.localize(DEFAULT_LOCALIZER))):
                portable = self.sut_cls.data().porter.dump(sample.subject)
                loaded = self.sut_cls.data().porter.load(portable)
                dumped = self.sut_cls.data().porter.dump(loaded)
                assert self.sut_cls.data().porter.dump(loaded) == dumped, (
                    f'Failed asserting that repeatedly loading and dumping sample "{sample.label.localize(DEFAULT_LOCALIZER)}" keeps producing the same portable data'
                )
                for other_sample in samples:
                    if other_sample is sample:
                        continue
                    assert (
                        self.sut_cls.data().porter.dump(other_sample.subject) != dumped
                    ), (
                        f'Failed asserting that sample "{sample.label.localize(DEFAULT_LOCALIZER)}" instance is not equal to sample "{other_sample.label.localize(DEFAULT_LOCALIZER)}"'
                    )

    @pytest.mark.parametrize(
        "other",
        [
            object,
            object(),
            True,
            False,
            None,
            123,
            "abc",
            [],
            {},
        ],
    )
    def test___eq____with_non_data_other(
        self, other: Any, subtests: pytest.Subtests
    ) -> None:
        """
        Tests :py:meth:`object.__eq__` implementations with values that do not subclass :py:class:`betty.data.Data`.
        """
        samples = list(self.sut_cls.data().samples)
        for sample in samples:
            with subtests.test(str(sample.label.localize(DEFAULT_LOCALIZER))):
                assert sample.subject != other

    def test___eq____with_samples(self, subtests: pytest.Subtests) -> None:
        """
        Tests :py:meth:`object.__eq__` implementations with the data definition's samples.
        """
        samples = list(self.sut_cls.data().samples)
        for sample in samples:
            with subtests.test(str(sample.label.localize(DEFAULT_LOCALIZER))):
                assert sample.subject == sample.subject, (
                    f'Failed asserting that sample "{sample.label.localize(DEFAULT_LOCALIZER)}" instance is equal to itself'
                )
                for other_sample in samples:
                    if other_sample is sample:
                        continue
                    assert sample.subject != other_sample.subject, (
                        f'Failed asserting that sample "{sample.label.localize(DEFAULT_LOCALIZER)}" instance is not equal to sample "{other_sample.label.localize(DEFAULT_LOCALIZER)}"'
                    )


@ObjectDefinition(label=Plain("Dummy data"))
class DummyData(Data):
    """
    A dummy :py:class:`betty.data.Data` implementation.
    """

    value = Optional(Property(StrDefinition(label="Value")))

    def __init__(self, /, value: str | None = None):
        self.value = value



from __future__ import annotations  # noqa: D100

import asyncio
from typing import TYPE_CHECKING, Self, final, override

from betty import documentation
from betty.console.command import Command, CommandDefinition, CommandFunction
from betty.locale.localizable.gettext import _
from betty.service.factory import Manufacturable
from betty.service.requirement.app import require_app

if TYPE_CHECKING:
    import argparse

    from betty.app import App


@final
@CommandDefinition(
    "docs",
    label=_("View the documentation"),
    description=_(
        "View Betty's interactive documentation. This will open your web browser."
    ),
)
class Docs(Manufacturable, Command):
    """
    .. plugin:: command:docs.
    """

    def __init__(self, app: App):
        self._app = app

    @override
    @classmethod
    @require_app
    async def new(cls, app: App, /) -> Self:
        return cls(app)

    @override
    async def configure(self, parser: argparse.ArgumentParser) -> CommandFunction:
        return self._command_function

    async def _command_function(self) -> None:
        server = documentation.DocumentationServer(
            self._app.binary_file_cache.with_scope("documentation").path,
            user=self._app.user,
        )
        async with server:
            await server.show()
            while True:
                await asyncio.sleep(999)
